"""Unit tests for the structure class."""  # noqa: INP001

from typing import Any

import pytest
from context import (
    GeckoAsyncStructure,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
)


class MockStructure(GeckoAsyncStructure):
    """Wrapper to help testing."""

    last_pos: int | None = None
    last_len: int | None = None
    last_value: Any | None = None

    def __init__(self) -> None:
        """Intialize the mock structure."""
        super().__init__(self.set_value_cb)

        class CfgClass:
            """The log class."""

            def __init__(self, struct_: GeckoAsyncStructure) -> None:
                self.struct = struct_

            @property
            def temperature_keys(self) -> list:
                return []

            @property
            def output_keys(self) -> list:
                return []

            @property
            def accessors(self) -> dict:
                return {
                    # <QuietState Type="Enum" Pos="362" Items="NOT_SET|QUIET|DRAIN|SOAK|OFF" RW="ALL" />  # noqa: E501
                    "QuietState": GeckoEnumStructAccessor(
                        self.struct,
                        "QuietState",
                        362,
                        None,
                        "NOT_SET|QUIET|DRAIN|SOAK|OFF",
                        None,
                        None,
                        "All",
                    ),
                    # <UdP1 Type="Enum" Pos="275" BitPos="14" Size="2" MaxItems="4" Items="OFF|LO|HI" RW="ALL" />  # noqa: E501
                    "UdP1": GeckoEnumStructAccessor(
                        self.struct, "UdP1", 275, 14, "OFF|LO|HI", 2, 4, "All"
                    ),
                    # <UdP2 Type="Enum" Pos="275" BitPos="12" Size="2" MaxItems="4" Items="OFF|LO|HI" RW="ALL" />  # noqa: E501
                    "UdP2": GeckoEnumStructAccessor(
                        self.struct, "UdP2", 275, 12, "OFF|LO|HI", 2, 4, "All"
                    ),
                    # <UdP3 Type="Enum" Pos="275" BitPos="11" Size="2" MaxItems="2" Items="OFF|HI" RW="ALL" />  # noqa: E501
                    "UdP3": GeckoEnumStructAccessor(
                        self.struct, "UdP3", 275, 11, "OFF|HI", 2, 2, "All"
                    ),
                    # <UdP4 Type="Enum" Pos="275" BitPos="10" Size="2" MaxItems="2" Items="OFF|HI" RW="ALL" />  # noqa: E501
                    "UdP4": GeckoEnumStructAccessor(
                        self.struct, "UdP4", 275, 10, "OFF|HI", 2, 2, "All"
                    ),
                    # <UdP5 Type="Enum" Pos="275" BitPos="9" Size="2" MaxItems="2" Items="OFF|HI" RW="ALL" />  # noqa: E501
                    "UdP5": GeckoEnumStructAccessor(
                        self.struct, "UdP5", 275, 9, "OFF|HI", 2, 2, "All"
                    ),
                    # <UdBL Type="Enum" Pos="275" BitPos="8" Size="2" MaxItems="2" Items="OFF|ON" RW="ALL" />  # noqa: E501
                    "UdBL": GeckoEnumStructAccessor(
                        self.struct, "UdBL", 275, 8, "OFF|ON", 2, 2, "All"
                    ),
                    # <UdL120 Type="Enum" Pos="275" BitPos="1" Size="2" MaxItems="2" Items="OFF|ON" RW="ALL" />  # noqa: E501
                    "UdL120": GeckoEnumStructAccessor(
                        self.struct, "UdL120", 275, 1, "OFF|ON", 2, 2, "All"
                    ),
                    # <UdLi Type="Enum" Pos="306" BitPos="0" MaxItems="4" Items="OFF|LO|MED|HI" RW="ALL" />  # noqa: E501
                    "UdLi": GeckoEnumStructAccessor(
                        self.struct, "UdLi", 306, 0, "OFF|LO|MED|HI", None, 4, "All"
                    ),
                }

        class LogClass:
            def __init__(self, struct_: GeckoAsyncStructure) -> None:
                self.struct = struct_

            @property
            def output_keys(self) -> list:
                return []

            @property
            def temperature_keys(self) -> list:
                return []

            @property
            def all_device_keys(self) -> list:
                return []

            @property
            def user_demand_keys(self) -> list:
                return []

            @property
            def error_keys(self) -> list:
                return []

            @property
            def accessors(self) -> dict:
                return {
                    # <UdValve Type="Enum" Pos="275" BitPos="7" Size="2" MaxItems="2" Items="OFF|ON" RW="ALL" />  # noqa: E501
                    # <UdTvLift Type="Enum" Pos="275" BitPos="5" Size="2" MaxItems="2" Items="OFF|ON" RW="ALL" />  # noqa: E501
                    # <UdSpkrLift Type="Enum" Pos="275" BitPos="4" Size="2" MaxItems="2" Items="OFF|ON" RW="ALL" />  # noqa: E501
                    # <UdVSP1 Type="Byte" Pos="303" Min="0" Max="100" RW="ALL" />
                    "UdVSP1": GeckoByteStructAccessor(
                        self.struct, "UdVSP1", 303, "All"
                    ),
                    # <UdVSP3 Type="Byte" Pos="304" Min="0" Max="100" RW="ALL" />
                    # <UdPumpTime Type="Byte" Pos="358" RW="ALL" />
                    # <UdQuietTime Type="Byte" Pos="359" Max="90" RW="ALL" />
                    # <UdLightTime Type="Byte" Pos="360" RW="ALL" />
                    # <UdL120Time Type="Byte" Pos="361" RW="ALL" />
                }

        self.config_class = CfgClass(self)
        self.log_class = LogClass(self)
        self.build_accessors()

    async def set_value_cb(self, pos: int, len_: int, newvalue: Any) -> None:
        self.last_pos = pos
        self.last_len = len_
        self.last_value = newvalue


class TestStruct:
    """Test the GeckoStruct class."""

    struct = MockStructure()

    def test_constructor(self) -> None:
        assert self.struct.status_block == b"\x00" * 1024
        assert len(self.struct.accessors) == 10
        assert self.struct.accessors["UdL120"].value == "OFF"

    @pytest.mark.asyncio
    async def test_set_value_callback(self) -> None:
        await self.struct.accessors["UdP1"].async_set_value("HI")
        assert self.struct.last_pos == 275
        assert self.struct.last_len == 2
        assert self.struct.last_value == 32768

    def test_observer_callback(self) -> None:
        p2_info = {"oldvalue": None, "newvalue": None}

        def p2_watch(_sender: Any, oldvalue: Any, newvalue: Any) -> None:
            p2_info["oldvalue"] = oldvalue
            p2_info["newvalue"] = newvalue

        self.struct.accessors["UdP2"].watch(p2_watch)
        self.struct.replace_status_block_segment(275, b"\x20\x00")
        assert p2_info["newvalue"] == "HI"



"""GeckoPack - A class to manage the pack for 'InXE'."""

from . import GeckoAsyncStructure


class GeckoPack:
    """A GeckoPack class for a specific spa."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the GeckoPack class."""
        self.struct = struct_

    @property
    def name(self) -> str:
        """Get the plateform name."""
        return "InXE"

    @property
    def plateform_type(self) -> int:
        """Get the plateform type."""
        return 1

    @property
    def plateform_segment(self) -> str:
        """Get the plateform segment."""
        return "aMainControl"

    @property
    def revision(self) -> str:
        """Get the SpaPackStruct revision."""
        return "39.0"



"""Connection events issued during spa discovery/connection."""

from __future__ import annotations

from collections.abc import Callable, Coroutine
from enum import Enum
from typing import Any


class GeckoSpaEvent(Enum):
    """Events triggered during locating and connection status of spas."""

    # Spa manager -------------------------------------------------------
    SPA_MAN_ENTER = 1
    """Spa manager was initialized"""
    SPA_MAN_EXIT = 2
    """Spa manager was exited, exc_info= exception if present"""

    SPA_NOT_FOUND = 5
    """Specified spa was not found on the network, spa_address= and
    spa_identifier= have more information"""

    # Locating spas -----------------------------------------------------
    LOCATING_STARTED = 100
    """Locating spas on the network has started"""

    LOCATING_DISCOVERED_SPA = 150
    """A spa was located, spa_descriptor= contains the details """

    LOCATING_FINISHED = 199
    """Locating spas on the network has finished, spa_descriptors=
    contains a list (possibly empty) of the spas that were found"""

    # Connecting to spas ------------------------------------------------
    CONNECTION_STARTED = 200
    """Connecting to a spa on the network has started"""

    CONNECTION_GOT_FIRMWARE_VERSION = 201
    """Connection has in.touch2 firmware versions, **kwargs has details"""

    CONNECTION_GOT_CHANNEL = 202
    """Connection has channel and RF strength. **kwargs has details"""

    CONNECTION_GOT_CONFIG_FILES = 203
    """Connection has got config files and versions. **kwargs has details"""

    CONNECTION_INITIAL_DATA_BLOCK_REQUEST = 204
    """Connection has asked for initial data block"""

    CONNECTION_SPA_COMPLETE = 250
    """Connection has completed for this spa, the facade can be constructed"""

    CONNECTION_CANNOT_FIND_LOG_VERSION = 295
    """During connection, the log version of the spa pack count not be
    found. **kwargs has details"""

    CONNECTION_CANNOT_FIND_CONFIG_VERSION = 296
    """During connection, the config version of the spa pack count not be
    found. **kwargs has details"""

    CONNECTION_CANNOT_FIND_SPA_PACK = 297
    """During connection, the calculated spa pack could not be found.
    **kwargs has details"""

    CONNECTION_PROTOCOL_RETRY_TIME_EXCEEDED = 298
    """During connection, the retry time was exceeded for part of the
    protocol handshake"""

    CONNECTION_FINISHED = 299
    """Connecting to a spa on the network has finished, facade=
    has the connected facade or None if some error occurred"""

    # Running spa ------------------------------------------------------
    RUNNING_PING_RECEIVED = 300
    """Running spa received a ping response"""
    RUNNING_PING_MISSED = 301
    """Running spa missed a ping response"""
    RUNNING_PING_NO_RESPONSE = 302
    """Running spa not responding to pings """
    RUNNING_SPA_DISCONNECTED = 303
    """Running spa was disconnected"""
    RUNNING_SPA_WATER_CARE_ERROR = 304
    """A running spa will sometimes get an unprompted watercare error"""
    RUNNING_SPA_PACK_REFRESHED = 305
    """A running spa will update the spapack data and radio strength periodically"""
    RUNNING_SPA_NEEDS_RELOAD = 306
    """A running spa has detected that it needs a complete reload"""

    # Events targeted at clients, to determine when things can be shown
    # or hidden in the UI
    CLIENT_HAS_STATUS_SENSOR = 401
    """The spa manager has the status sensor"""
    CLIENT_HAS_RECONNECT_BUTTON = 402
    """The spa manager has the reconnect button"""
    CLIENT_HAS_PING_SENSOR = 403
    """The spa manager has the ping sensor"""

    CLIENT_FACADE_IS_READY = 420
    """The facade is ready for use."""
    CLIENT_FACADE_TEARDOWN = 421
    """Facade is being torn down."""

    # Error event that SpaMan can retry without ------------------------
    # intervention

    # Terminal states that require subclasses to ------------------------
    # let go of the facade and all automation objects and reconnect
    ERROR_TOO_MANY_RF_ERRORS = 500
    """The spa has had too many RF errors"""

    ERROR_PROTOCOL_RETRY_TIME_EXCEEDED = 501
    """Protocol retry time was exceeded during normal operations"""

    ERROR_RF_ERROR = 502
    """The in.touch EN module can't communicate with the CO module"""

    CallBack = Callable[..., Coroutine[Any, Any, None]]
    """Typedef for event callbacks, located here so we can be DRY"""



"""Gecko REQRM/RMREQ handlers."""

from __future__ import annotations

import logging
import struct
from enum import IntEnum
from typing import Any

from geckolib.config import GeckoConfig

from .packet import GeckoPacketProtocolHandler

REQRM_VERB = b"REQRM"  # Request all reminders
RMREQ_VERB = b"RMREQ"  # Response with all 10 reminders
SETRM_VERB = b"SETRM"  # Set all 10 reminders
RMSET_VERB = b"RMSET"  # Ack for the reminder set

RESPONSE_FORMAT = ">BBB"

_LOGGER = logging.getLogger(__name__)


class GeckoReminderType(IntEnum):
    """Reminder type class."""

    INVALID = 0
    RINSE_FILTER = 1
    CLEAN_FILTER = 2
    CHANGE_WATER = 3
    CHECK_SPA = 4
    CHANGE_OZONATOR = 5
    CHANGE_VISION_CARTRIDGE = 6

    @staticmethod
    def to_string(the_type: GeckoReminderType) -> str:  # noqa: PLR0911
        """Converet enum to string."""
        if the_type == GeckoReminderType.INVALID:
            return "Invalid"
        if the_type == GeckoReminderType.RINSE_FILTER:
            return "RinseFilter"
        if the_type == GeckoReminderType.CLEAN_FILTER:
            return "CleanFilter"
        if the_type == GeckoReminderType.CHANGE_WATER:
            return "ChangeWater"
        if the_type == GeckoReminderType.CHECK_SPA:
            return "CheckSpa"
        if the_type == GeckoReminderType.CHANGE_OZONATOR:
            return "ChangeOzonator"
        if the_type == GeckoReminderType.CHANGE_VISION_CARTRIDGE:
            return "ChangeVisionCartridge"
        # Technically unreachable code here
        return "Unhandled"


class GeckoRemindersProtocolHandler(GeckoPacketProtocolHandler):
    """Reminders protocol handler."""

    @staticmethod
    def request(seq: int, **kwargs: Any) -> GeckoRemindersProtocolHandler:
        """Generate request."""
        return GeckoRemindersProtocolHandler(
            content=b"".join([REQRM_VERB, struct.pack(">B", seq)]),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS,
            on_retry_failed=GeckoPacketProtocolHandler.default_retry_failed_handler,
            **kwargs,
        )

    @staticmethod
    def req_response(
        reminders: list[tuple[GeckoReminderType, int]], **kwargs: Any
    ) -> GeckoRemindersProtocolHandler:
        """Generate response handler."""
        return GeckoRemindersProtocolHandler(
            content=b"".join(
                [RMREQ_VERB]
                + [
                    struct.pack("<BhB", reminder[0], reminder[1], 1)
                    for reminder in reminders
                ]
            ),
            **kwargs,
        )

    @staticmethod
    def set(
        seq: int, reminders: list[tuple[GeckoReminderType, int]], **kwargs: Any
    ) -> GeckoRemindersProtocolHandler:
        """Generate a set command."""
        return GeckoRemindersProtocolHandler(
            content=b"".join(
                [SETRM_VERB, struct.pack(">B", seq)]
                + [
                    struct.pack("<BhB", reminder[0], reminder[1], 1)
                    for reminder in reminders
                ]
            ),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS,
            on_retry_failed=GeckoPacketProtocolHandler.default_retry_failed_handler,
            **kwargs,
        )

    @staticmethod
    def set_ack(**kwargs: Any) -> GeckoRemindersProtocolHandler:
        """Generate a set response."""
        return GeckoRemindersProtocolHandler(content=b"".join([RMSET_VERB]), **kwargs)

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the reminders protocol handler class."""
        super().__init__(**kwargs)
        self.reminders: list[tuple[GeckoReminderType, int]] = []
        self.is_request: bool = False

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle this verb."""
        return received_bytes.startswith(
            (REQRM_VERB, RMREQ_VERB, SETRM_VERB, RMSET_VERB)
        )

    def _extract_reminders(self, remainder: bytes) -> None:
        rest = remainder
        while len(rest) > 0:
            (t, days, _push, rest) = struct.unpack(f"<BhB{len(rest) - 4}s", rest)
            try:
                self.reminders.append((GeckoReminderType(t), days))
            except ValueError:
                _LOGGER.warning("Cannot use %d as reminder type, ignored", t)

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the verb."""
        remainder = received_bytes[5:]
        self.is_request = False
        self.reminders = []

        if received_bytes.startswith(REQRM_VERB):
            self._sequence = struct.unpack(">B", remainder[0:1])[0]
            self.is_request = True
            return  # Stay in the handler list

        if received_bytes.startswith(SETRM_VERB):
            self._sequence = struct.unpack(">B", remainder[0:1])[0]
            self._extract_reminders(remainder[1:])
            return  # Stay in the handler list

        if received_bytes.startswith(RMSET_VERB):
            pass

        # Otherwise must be RMREQ
        if received_bytes.startswith(RMREQ_VERB):
            self._extract_reminders(remainder)

        self._should_remove_handler = True



"""GeckoLogStruct - A class to manage the LogStruct for 'InXM v5'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 5

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 256

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 479

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return [
            "P1",
            "P2",
            "P3",
            "P4",
            "P5",
            "BL",
            "CP",
            "O3",
            "L120",
            "FAN",
            "Fb",
            "AlwayOn",
            "TvLift",
            "SpkrLift",
            "Sani",
            "Onzen",
            "RhDuty",
            "Vsp1Front",
            "Vsp1Min",
            "VSP3Front",
            "LI",
        ]

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return [
            "UdP1",
            "UdP2",
            "UdP3",
            "UdP4",
            "UdP5",
            "UdBL",
            "UdTvLift",
            "UdSpkrLift",
            "UdFb",
            "UdL120",
            "UdQuiet",
            "UdLi",
            "UdVSP1",
            "UdVSP3",
            "UdPumpTime",
            "UdQuietTime",
            "UdLightTime",
            "UdL120Time",
            "UdDrain",
        ]

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return [
            "AmbiantOHLevel1",
            "AmbiantOHLevel2",
            "ContactorCoilFailErr",
            "FreqDetecErr",
            "Fuse1Err",
            "Fuse2Err",
            "Fuse3Err",
            "KinPumpOff",
            "LearnPhaseErr",
            "NoHeaterCurrentErr",
            "RegOverHeat",
            "RelayStuck",
            "RhCommErr",
            "RhDutyErr",
            "RhRegProbeErr",
            "ScanErr",
            "SupplyErr",
            "TempNotValid",
            "ThermistanceErr",
            "rHId",
        ]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "RhWaterTemp": GeckoTempStructAccessor(
                self.struct, "LogStructure/RealTimeTemp/RhWaterTemp", 261, None
            ),
            "RhTriacTemp": GeckoTempStructAccessor(
                self.struct, "LogStructure/RealTimeTemp/RhTriacTemp", 263, None
            ),
            "Hours": GeckoByteStructAccessor(
                self.struct, "LogStructure/MiscStatus/Hours", 256, None
            ),
            "Menu": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/MiscStatus/Menu",
                353,
                None,
                [
                    "NORMAL",
                    "S1",
                    "S2",
                    "H1",
                    "H2",
                    "H3",
                    "H4",
                    "Fr",
                    "LearnFail",
                    "Learning",
                    "BreakerSelect",
                    "PhaseSelect",
                    "ConfigSelect",
                    "StickBank",
                    "K600Upload",
                    "Powerup",
                ],
                None,
                None,
                None,
            ),
            "HeaterPhaseSign": GeckoByteStructAccessor(
                self.struct, "LogStructure/MiscStatus/HeaterPhaseSign", 300, None
            ),
            "OUT1ARead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT1ARead", 280, None
            ),
            "OUT1BRead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT1BRead", 282, None
            ),
            "OUT2ARead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT2ARead", 284, None
            ),
            "OUT2BRead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT2BRead", 286, None
            ),
            "OUT3ARead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT3ARead", 288, None
            ),
            "OUT4ARead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT4ARead", 290, None
            ),
            "OUT5ARead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT5ARead", 292, None
            ),
            "OUT6ARead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT6ARead", 294, None
            ),
            "OUT7ARead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/OUT7ARead", 296, None
            ),
            "HeaterRead": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/HeaterRead", 298, None
            ),
            "UdP1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP1",
                275,
                14,
                ["OFF", "LO", "HI"],
                2,
                4,
                "ALL",
            ),
            "UdP2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP2",
                275,
                12,
                ["OFF", "LO", "HI"],
                2,
                4,
                "ALL",
            ),
            "UdP3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP3",
                275,
                11,
                ["OFF", "HI"],
                2,
                2,
                "ALL",
            ),
            "UdP4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP4",
                275,
                10,
                ["OFF", "HI"],
                2,
                2,
                "ALL",
            ),
            "UdP5": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP5",
                275,
                9,
                ["OFF", "HI"],
                2,
                2,
                "ALL",
            ),
            "UdBL": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdBL",
                275,
                8,
                ["OFF", "ON"],
                2,
                2,
                "ALL",
            ),
            "UdTvLift": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdTvLift",
                275,
                5,
                ["OFF", "ON"],
                2,
                2,
                "ALL",
            ),
            "UdSpkrLift": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdSpkrLift",
                275,
                4,
                ["OFF", "ON"],
                2,
                2,
                "ALL",
            ),
            "UdFb": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdFb",
                275,
                2,
                ["OFF", "FIX", "NA", "SCAN"],
                2,
                4,
                "ALL",
            ),
            "UdL120": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdL120",
                275,
                1,
                ["OFF", "ON"],
                2,
                2,
                "ALL",
            ),
            "UdQuiet": GeckoBoolStructAccessor(
                self.struct, "LogStructure/UserDemands/UdQuiet", 275, 0, "ALL"
            ),
            "UdLi": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdLi",
                306,
                0,
                ["OFF", "LO", "MED", "HI"],
                None,
                4,
                "ALL",
            ),
            "UdVSP1": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdVSP1", 303, "ALL"
            ),
            "UdVSP3": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdVSP3", 304, "ALL"
            ),
            "UdPumpTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdPumpTime", 358, "ALL"
            ),
            "UdQuietTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdQuietTime", 359, "ALL"
            ),
            "UdLightTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdLightTime", 360, "ALL"
            ),
            "UdL120Time": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdL120Time", 361, "ALL"
            ),
            "UdDrain": GeckoBoolStructAccessor(
                self.struct, "LogStructure/UserDemands/UdDrain", 362, None, "ALL"
            ),
            "P1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P1",
                356,
                0,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P2",
                356,
                2,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P3",
                356,
                4,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P4",
                356,
                6,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P5": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P5",
                355,
                0,
                ["OFF", "HIGH"],
                None,
                2,
                None,
            ),
            "BL": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/BL",
                355,
                1,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "CP": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/CP",
                355,
                2,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "O3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/O3",
                355,
                3,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "L120": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/L120",
                355,
                4,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "FAN": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/FAN",
                355,
                7,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "Fb": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/Fb",
                354,
                0,
                ["OFF", "OFF", "FIX", "SCAN"],
                None,
                4,
                None,
            ),
            "AlwayOn": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/AlwayOn",
                354,
                2,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "TvLift": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/TvLift",
                354,
                3,
                ["DOWN", "UP"],
                None,
                2,
                None,
            ),
            "SpkrLift": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/SpkrLift",
                354,
                4,
                ["DOWN", "UP"],
                None,
                2,
                None,
            ),
            "Sani": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/Sani",
                354,
                5,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "Onzen": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/Onzen",
                354,
                6,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "RhDuty": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/RhDuty", 274, None
            ),
            "Vsp1Front": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/Vsp1Front", 301, None
            ),
            "Vsp1Min": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/Vsp1Min", 305, None
            ),
            "VSP3Front": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/VSP3Front", 302, None
            ),
            "FilterAccess": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/FilterAccess",
                348,
                0,
                ["INTERNAL", "REMOTE"],
                None,
                2,
                "ALL",
            ),
            "RemoteFiltAction": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteFiltAction",
                307,
                None,
                ["IDLE", "STOP", "START", "NEW", "ACTIVE"],
                None,
                None,
                "ALL",
            ),
            "RemoteFiltDur": GeckoTimeStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteFiltDur", 308, "ALL"
            ),
            "RemoteFiltDurPerDay": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteFiltDurPerDay",
                316,
                "ALL",
            ),
            "OnzenAccess": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/OnzenAccess",
                348,
                1,
                ["INTERNAL", "REMOTE"],
                None,
                2,
                "ALL",
            ),
            "RemoteOnzenAction": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteOnzenAction",
                310,
                None,
                ["IDLE", "STOP", "START", "NEW", "ACTIVE"],
                None,
                None,
                "ALL",
            ),
            "RemoteOnzenDur": GeckoTimeStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteOnzenDur", 311, "ALL"
            ),
            "EconomyAccess": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/EconomyAccess",
                348,
                2,
                ["INTERNAL", "REMOTE"],
                None,
                2,
                "ALL",
            ),
            "RemoteEconAction": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteEconAction",
                313,
                None,
                ["IDLE", "STOP", "START", "NEW", "ACTIVE"],
                None,
                None,
                "ALL",
            ),
            "RemoteEconDur": GeckoTimeStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteEconDur", 314, "ALL"
            ),
            "RemoteConfigIndex": GeckoByteStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteConfigIndex", 350, "ALL"
            ),
            "RemoteNbOfPhases": GeckoByteStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteNbOfPhases", 351, "ALL"
            ),
            "RemoteBreakerIndex": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteBreakerIndex",
                352,
                "ALL",
            ),
            "Purge": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/Purge", 269, 7, None
            ),
            "Clean": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/Clean", 269, 5, None
            ),
            "CleanSuspended": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/CleanSuspended", 269, 4, None
            ),
            "CleanSuspendUD": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/CleanSuspendUD", 279, 3, None
            ),
            "FiltOT": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/FiltOT", 279, 0, None
            ),
            "CPOT": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/CPOT", 279, 1, None
            ),
            "Drainning": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/Drainning", 279, 7, None
            ),
            "SwmActive": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SwmActive", 267, None, None
            ),
            "SwmRisk": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SWMStatus/SwmRisk",
                268,
                None,
                ["NO", "LO", "MED", "HI", "EXTREME"],
                None,
                None,
                None,
            ),
            "SwmPurge": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SwmPurge", 269, 3, None
            ),
            "SwmPurgeSusp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SwmPurgeSusp", 269, 2, None
            ),
            "RealSetPointG": GeckoTempStructAccessor(
                self.struct, "LogStructure/RegulationStatus/RealSetPointG", 272, None
            ),
            "DisplayedTempG": GeckoTempStructAccessor(
                self.struct, "LogStructure/RegulationStatus/DisplayedTempG", 265, None
            ),
            "TempNotValid": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/TempNotValid", 349, 1, None
            ),
            "PumpRun": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/PumpRun", 269, 1, None
            ),
            "CheckFlo": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/CheckFlo", 269, 0, None
            ),
            "Heating": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/Heating", 270, 7, None
            ),
            "CoolingDown": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/CoolingDown", 270, 6, None
            ),
            "KinPurgeActive": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/RegulationStatus/KinPurgeActive",
                270,
                5,
                None,
            ),
            "EconActive": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/EconActive", 271, 0, None
            ),
            "ProgEconActive": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/RegulationStatus/ProgEconActive",
                271,
                2,
                None,
            ),
            "FreqDetecErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/FreqDetecErr", 277, 6, None
            ),
            "ThermistanceErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/ThermistanceErr", 277, 5, None
            ),
            "ContactorCoilFailErr": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/ErrorMessages/ContactorCoilFailErr",
                277,
                4,
                None,
            ),
            "NoHeaterCurrentErr": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/ErrorMessages/NoHeaterCurrentErr",
                277,
                3,
                None,
            ),
            "LearnPhaseErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/LearnPhaseErr", 277, 2, None
            ),
            "Fuse3Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/Fuse3Err", 277, 1, None
            ),
            "Fuse2Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/Fuse2Err", 277, 0, None
            ),
            "Fuse1Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/Fuse1Err", 278, 7, None
            ),
            "SupplyErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SupplyErr", 278, 6, None
            ),
            "ScanErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/ScanErr", 278, 5, None
            ),
            "RegOverHeat": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/RegOverHeat", 278, 4, None
            ),
            "AmbiantOHLevel2": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/AmbiantOHLevel2", 278, 3, None
            ),
            "AmbiantOHLevel1": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/AmbiantOHLevel1", 278, 2, None
            ),
            "RelayStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/RelayStuck", 278, 1, None
            ),
            "rHId": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/rHId", 278, 0, None
            ),
            "KinPumpOff": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/KinPumpOff", 279, 4, None
            ),
            "RhCommErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhCommErr", 257, 7, None
            ),
            "RhNotEnoughHeat": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhNotEnoughHeat", 257, 3, None
            ),
            "RhRegSlope": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhRegSlope", 257, 2, None
            ),
            "RhLowFlo": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhLowFlo", 257, 0, None
            ),
            "RhDutyErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhDutyErr", 258, 7, None
            ),
            "RhNoFloRetry": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhNoFloRetry", 258, 6, None
            ),
            "RhNoFloTemp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhNoFloTemp", 258, 5, None
            ),
            "RhNoFloXTries": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhNoFloXTries", 258, 4, None
            ),
            "RhHrHL": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHrHL", 258, 3, None
            ),
            "RhHwHL": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHwHL", 258, 2, None
            ),
            "RhRegProbeErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhRegProbeErr", 258, 1, None
            ),
            "RhRegOverHeat": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhRegOverHeat", 258, 0, None
            ),
            "RhHrTriacPr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHrTriacPr", 259, 7, None
            ),
            "RhHrTriacOH": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHrTriacOH", 259, 6, None
            ),
            "RhSWTriacOH": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhSWTriacOH", 259, 5, None
            ),
            "RhHwTriacOH": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHwTriacOH", 259, 4, None
            ),
            "RhHrKin": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHrKin", 259, 3, None
            ),
            "RhHrKinNoFlo": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHrKinNoFlo", 259, 2, None
            ),
            "RhSwKinTemp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhSwKinTemp", 259, 1, None
            ),
            "RhHWKin": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHWKin", 259, 0, None
            ),
            "RhHeaterDisabled": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHeaterDisabled", 260, 4, None
            ),
            "RhFloDetected": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhFloDetected", 260, 3, None
            ),
            "RhFloCheck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhFloCheck", 260, 2, None
            ),
            "RhHeaterSuspended": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/InthermStatus/RhHeaterSuspended",
                260,
                1,
                None,
            ),
            "RhFloCheckReady": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhFloCheckReady", 260, 0, None
            ),
            "PackBootID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootID", 317, None
            ),
            "PackBootRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootRev", 319, None
            ),
            "PackBootRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootRel", 320, None
            ),
            "PackType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackType",
                321,
                None,
                [
                    "Unknown",
                    "inXE",
                    "MasIBC",
                    "MIA",
                    "DJS4",
                    "inClear",
                    "inXM",
                    "K600",
                    "inTerface",
                    "inTouch",
                ],
                None,
                None,
                None,
            ),
            "PackMemRange": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackMemRange",
                322,
                0,
                ["16K", "32K", "48K", "64K"],
                None,
                4,
                None,
            ),
            "PackRegion": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackRegion",
                322,
                2,
                ["UL", "CE"],
                None,
                2,
                None,
            ),
            "PackStatusLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackStatusLib", 323, None
            ),
            "PackConfigLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfigLib", 324, None
            ),
            "PackCoreID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreID", 325, None
            ),
            "PackCoreRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRev", 327, None
            ),
            "PackCoreRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRel", 328, None
            ),
            "PackConfID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfID", 329, None
            ),
            "PackConfRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfRev", 331, None
            ),
            "PackConfRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfRel", 332, None
            ),
            "PackNumberOfConf": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackNumberOfConf", 333, None
            ),
            "PackLogTrig": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackLogTrig",
                357,
                None,
                ["Restricted", "Full"],
                None,
                2,
                "ALL",
            ),
            "inThermID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/inThermID", 335, None
            ),
            "inThermRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/inThermRev", 337, None
            ),
            "inThermRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/inThermRel", 338, None
            ),
            "inThermCur1KW": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/inThermCur1KW", 339, None
            ),
            "inThermCurFull": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/inThermCurFull", 341, None
            ),
            "K600ID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/K600ID", 343, None
            ),
            "K600Rev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/K600Rev", 345, None
            ),
            "K600Rel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/K600Rel", 346, None
            ),
            "K600Protocol": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/K600Protocol", 347, None
            ),
        }



"""GeckoLogStruct - A class to manage the LogStruct for 'MAS-IBC-32K v1'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 1

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 256

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 285

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return [
            "KeypadID",
            "KeypadRev",
            "BlowerState",
            "PurgeStart",
            "PurgeDelayTimer",
            "PurgeStandby",
            "PowerState",
            "KeypadLock",
            "LI",
        ]

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return []

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "UserBlowerIntensity": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UserBlowerIntensity", 256, "ALL"
            ),
            "UserChroma": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UserChroma",
                257,
                None,
                [
                    "OFF",
                    "WHITE",
                    "RED",
                    "GREEN",
                    "BLUE",
                    "PURPLE",
                    "YELLOW",
                    "SCAN",
                    "ORANGE",
                    "PAUSE",
                ],
                None,
                None,
                "ALL",
            ),
            "UserBathTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UserBathTime", 258, "ALL"
            ),
            "UserHeater1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UserHeater1",
                259,
                None,
                ["OFF", "LOW", "MED", "HIGH"],
                None,
                None,
                "ALL",
            ),
            "UserHeater2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UserHeater2",
                260,
                None,
                ["OFF", "LOW", "MED", "HIGH"],
                None,
                None,
                "ALL",
            ),
            "UserGeysair": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UserGeysair",
                261,
                None,
                ["OFF", "ON"],
                None,
                None,
                None,
            ),
            "UserDryingCycle": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UserDryingCycle",
                262,
                0,
                ["AfterBath", "Daily"],
                None,
                2,
                "ALL",
            ),
            "UserDryingHour": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UserDryingHour", 263, "ALL"
            ),
            "UserDryingMinute": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UserDryingMinute", 264, "ALL"
            ),
            "UserDryingSecond": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UserDryingSecond", 265, "ALL"
            ),
            "UserDryingDelay": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UserDryingDelay",
                266,
                0,
                [
                    "0",
                    "",
                    "",
                    "",
                    "",
                    "5",
                    "",
                    "",
                    "",
                    "",
                    "10",
                    "",
                    "",
                    "",
                    "",
                    "15",
                    "",
                    "",
                    "",
                    "",
                    "20",
                    "",
                    "",
                    "",
                    "",
                    "25",
                    "",
                    "",
                    "",
                    "",
                    "30",
                    "",
                    "",
                    "",
                    "",
                    "35",
                    "",
                    "",
                    "",
                    "",
                    "40",
                    "",
                    "",
                    "",
                    "",
                    "45",
                    "",
                    "",
                    "",
                    "",
                    "50",
                    "",
                    "",
                    "",
                    "",
                    "55",
                    "",
                    "",
                    "",
                    "",
                    "60",
                ],
                None,
                63,
                "ALL",
            ),
            "KeypadID": GeckoWordStructAccessor(
                self.struct, "LogStructure/DeviceStatus/KeypadID", 267, None
            ),
            "KeypadRev": GeckoWordStructAccessor(
                self.struct, "LogStructure/DeviceStatus/KeypadRev", 269, None
            ),
            "BlowerState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/BlowerState",
                271,
                None,
                ["PAUSE", "RUN", "STATE10"],
                None,
                None,
                None,
            ),
            "PurgeStart": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/PurgeStart",
                262,
                1,
                ["DEACTIVATE", "ACTIVATE"],
                None,
                None,
                None,
            ),
            "PurgeDelayTimer": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/PurgeDelayTimer",
                262,
                2,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                    "16",
                    "17",
                    "18",
                    "19",
                    "20",
                    "21",
                    "22",
                    "23",
                    "24",
                    "25",
                    "26",
                    "27",
                    "28",
                    "29",
                    "30",
                    "31",
                    "32",
                    "33",
                    "34",
                    "35",
                    "36",
                    "37",
                    "38",
                    "39",
                    "40",
                    "41",
                    "42",
                    "43",
                    "44",
                    "45",
                    "46",
                    "47",
                    "48",
                    "49",
                    "50",
                    "51",
                    "52",
                    "53",
                    "54",
                    "55",
                    "56",
                    "57",
                    "58",
                    "59",
                    "60",
                ],
                None,
                None,
                None,
            ),
            "PurgeStandby": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/PurgeStandby",
                266,
                6,
                ["STOP", "START"],
                None,
                None,
                None,
            ),
            "PowerState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/PowerState",
                283,
                None,
                ["OFF", "ON"],
                None,
                None,
                None,
            ),
            "KeypadLock": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/KeypadLock",
                284,
                None,
                ["OFF", "ON"],
                None,
                None,
                None,
            ),
            "BootID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/BootID", 272, None
            ),
            "BootRev": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/BootRev", 274, None
            ),
            "PackCoreID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreID", 276, None
            ),
            "PackCoreRev": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRev", 278, None
            ),
            "PackCoreRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRel", 280, None
            ),
            "PackType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackType",
                281,
                None,
                [
                    "Unknown",
                    "inXE",
                    "MasIBC",
                    "MIA",
                    "DJS4",
                    "inClear",
                    "inXM",
                    "K600",
                    "inTerface",
                    "inTouch",
                    "inYT",
                ],
                None,
                None,
                None,
            ),
            "PackMemRange": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackMemRange",
                282,
                0,
                ["16K", "32K", "48K", "64K"],
                None,
                4,
                None,
            ),
        }



"""GeckoAsyncUdpProtocol - Gecko Async UDP protocol implementation."""

import asyncio
import logging
from collections.abc import Callable
from types import TracebackType
from typing import Any, TypeVar

from geckolib.async_taskman import GeckoAsyncTaskMan
from geckolib.config import GeckoConfig

from .async_peekablequeue import AsyncPeekableQueue
from .udp_protocol_handler import GeckoUdpProtocolHandler

_LOGGER = logging.getLogger(__name__)


class DbgLock(asyncio.Lock):
    """Class to debug locking semantics."""

    def __init__(self) -> None:
        """Initialize the class."""
        super().__init__()

    async def __aenter__(self) -> None:
        """Async with."""
        t = asyncio.current_task()
        _LOGGER.debug("LOCK: About to acquire for task `%s`", t.get_name())
        _LOGGER.debug(self)
        await super().__aenter__()
        _LOGGER.debug("LOCK: acquired for task `%s`", t.get_name())

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        """Async with."""
        t = asyncio.current_task()
        _LOGGER.debug("LOCK: About to release for task %s", t.get_name())
        await super().__aexit__(exc_type, exc, tb)
        _LOGGER.debug("LOCK: Released for task `%s`", t.get_name())
        _LOGGER.debug(self)


class GeckoAsyncUdpProtocol(asyncio.DatagramProtocol):
    """
    Gecko in.touch2 uses UDP to communicate.

    This class is an async UDP protocol handler for asyncio. It dispatches
    to classes derived from GeckoUdpProtocolHandler. Since it doesn't need
    to be thread safe it's a good deal more simple that its predecessor.
    """

    def __init__(
        self,
        taskman: GeckoAsyncTaskMan,
        on_connection_lost: asyncio.Event,
        destination: tuple[str, str],
    ) -> None:
        """Initialize the protocol class."""
        self.transport = None
        self._on_connection_lost: asyncio.Event = on_connection_lost
        self._destination = destination

        self._sequence_counter_protocol = 0
        self._sequence_counter_command = 191
        self._queue = AsyncPeekableQueue()
        self._taskman = taskman
        self._lock = asyncio.Lock()
        _LOGGER.debug("AsyncUdpProtocol started.")

    @property
    def lock(self) -> asyncio.Lock:
        """Get the lock object."""
        return self._lock

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Made connection."""
        _LOGGER.debug("GeckoAsyncUdpProtocol: connection made to %s.", transport)
        self.transport: asyncio.BaseTransport | None = transport
        self._on_connection_lost.clear()

    def connection_lost(self, exc: BaseException | None) -> None:
        """Lost connection."""
        _LOGGER.debug(
            "GeckoAsyncUdpProtocol: connection lost from %s (%s)", self.transport, exc
        )
        if self.transport is not None:
            self.transport.close()
            self.transport = None
        self._on_connection_lost.set()

    def error_received(self, exc: BaseException) -> None:
        """Error received."""
        _LOGGER.exception("GeckoAsyncUdpProtocol: Exception received %s", exc)

    @property
    def isopen(self) -> bool:
        """Check to see if the transport is connected."""
        return self.transport is not None

    def disconnect(self) -> None:
        """Disconnect the protocol from the socket."""
        if self.queue:
            _LOGGER.warning(
                "UDP protocol datagram queue not empty during disconnection",
            )
            while self.queue:
                item = self.queue.pop()
                _LOGGER.debug("Unprocessed[%s]", item)
        self.connection_lost(None)

    @property
    def queue(self) -> AsyncPeekableQueue:
        """Get the datagram queue."""
        return self._queue

    def queue_send(
        self,
        protocol_handler: GeckoUdpProtocolHandler,
        destination: tuple | None = None,
    ) -> None:
        """Queue a message to be sent async later."""
        if not self.isopen:
            _LOGGER.warning("Cannot queue message as transport is closed")
            return
        assert self.transport is not None  # noqa: S101
        if destination is None:
            destination = self._destination
        protocol_handler.last_destination = destination

        send_bytes = protocol_handler.send_bytes
        destination = (destination[0], destination[1])
        _LOGGER.debug("Sending %s to %s", send_bytes, destination)
        # transport.sendto is a non-blocking call.
        self.transport.sendto(send_bytes, destination)

    def get_and_increment_sequence_counter(self, *, command: bool = False) -> int:
        """Get (and increment) the sequence counter."""
        if command:
            if self._sequence_counter_command == 255:  # noqa: PLR2004
                self._sequence_counter_command = 191
            self._sequence_counter_command += 1
            return self._sequence_counter_command
        if self._sequence_counter_protocol == 191:  # noqa: PLR2004
            self._sequence_counter_protocol = 0
        self._sequence_counter_protocol += 1
        return self._sequence_counter_protocol

    def datagram_received(self, data: Any, addr: tuple) -> None:
        """Handle datagrame."""
        _LOGGER.debug("Datagram received: %s from %s", data, addr)
        if b"SPACK" in data or b"STATQ" in data:
            _LOGGER.debug("Put on command queue")
            self.queue.push_command((data, addr))
        else:
            self.queue.push((data, addr))

    T = TypeVar("T", bound="GeckoUdpProtocolHandler")

    async def get(
        self,
        create_func: Callable[[], T],
        destination: tuple | None = None,
        packet_timeout: int = GeckoConfig.PAUSE_BETWEEN_RETRIES_IN_SECONDS,
    ) -> T | None:
        """Get the response to the request."""
        async with self.lock:
            # Create the request
            request = create_func()

            try:
                async with asyncio.timeout(request.timeout_in_seconds):
                    while True:
                        # Queue it for delivery
                        self.queue_send(request, destination)

                        # Wait for a response up to a certain amount of time
                        if not await request.wait_for_response(self, packet_timeout):
                            # If handled, then return the handler which ought
                            # to contain the information as requested
                            return request

                        # Create a clean new request
                        request = create_func()

            except TimeoutError:
                _LOGGER.debug(
                    "TIMEOUT: Handler %s threw TimeoutError (%fs)",
                    request,
                    request.timeout_in_seconds,
                )



"""GeckoConfigStruct - A class to manage the ConfigStruct for 'InMix v4'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 4

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "InMix-Zone1Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone1Led",
                592,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "InMix-Zone1Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone1Type",
                592,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "InMix-Zone1Supply": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone1Supply",
                592,
                2,
                ["12V", "5V"],
                None,
                2,
                None,
            ),
            "InMix-Zone1UseOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone1UseOut1", 592, 4, None
            ),
            "InMix-Zone1UseOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone1UseOut2", 592, 5, None
            ),
            "InMix-Zone1UseOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone1UseOut3", 592, 6, None
            ),
            "InMix-Zone1UseOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone1UseOut4", 592, 7, None
            ),
            "InMix-Zone2Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone2Led",
                593,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "InMix-Zone2Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone2Type",
                593,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "InMix-Zone2Supply": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone2Supply",
                593,
                2,
                ["12V", "5V"],
                None,
                2,
                None,
            ),
            "InMix-Zone2UseOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone2UseOut1", 593, 4, None
            ),
            "InMix-Zone2UseOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone2UseOut2", 593, 5, None
            ),
            "InMix-Zone2UseOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone2UseOut3", 593, 6, None
            ),
            "InMix-Zone2UseOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone2UseOut4", 593, 7, None
            ),
            "InMix-Zone3Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone3Led",
                594,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "InMix-Zone3Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone3Type",
                594,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "InMix-Zone3Supply": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone3Supply",
                594,
                2,
                ["12V", "5V"],
                None,
                2,
                None,
            ),
            "InMix-Zone3UseOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone3UseOut1", 594, 4, None
            ),
            "InMix-Zone3UseOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone3UseOut2", 594, 5, None
            ),
            "InMix-Zone3UseOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone3UseOut3", 594, 6, None
            ),
            "InMix-Zone3UseOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone3UseOut4", 594, 7, None
            ),
            "InMix-Zone4Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone4Led",
                595,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "InMix-Zone4Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone4Type",
                595,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "InMix-Zone4Supply": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-Zone4Supply",
                595,
                2,
                ["12V", "5V"],
                None,
                2,
                None,
            ),
            "InMix-Zone4UseOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone4UseOut1", 595, 4, None
            ),
            "InMix-Zone4UseOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone4UseOut2", 595, 5, None
            ),
            "InMix-Zone4UseOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone4UseOut3", 595, 6, None
            ),
            "InMix-Zone4UseOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-Zone4UseOut4", 595, 7, None
            ),
            "InMix-NumberOfZones": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-NumberOfZones",
                596,
                0,
                ["0", "1", "2", "3", "4"],
                None,
                5,
                "ALL",
            ),
            "InMix-StatusLightNormalColor": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-StatusLightNormalColor",
                596,
                3,
                ["BLUE", "WHITE", "GREEN", "PURPLE", "YELLOW", "CYAN", "ORANGE", "OFF"],
                None,
                8,
                "ALL",
            ),
            "InMix-OutputAttribution": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/InMix-OutputAttribution", 596, 7, None
            ),
            "InMix-MinimumFadeIntensity": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-MinimumFadeIntensity",
                597,
                0,
                [
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                    "15",
                ],
                None,
                16,
                "ALL",
            ),
            "InMix-StatusLightRemindersColor": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InMix-StatusLightRemindersColor",
                597,
                4,
                [
                    "SAME_AS_NORMAL",
                    "WHITE",
                    "GREEN",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "OFF",
                ],
                None,
                8,
                "ALL",
            ),
            "InMix-RemindersNotBlinkingLED": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/InMix-RemindersNotBlinkingLED",
                597,
                7,
                None,
            ),
        }



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoVersionProtocolHandler,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoVersionHandler(unittest.TestCase):
    """Version handler tests."""

    def test_send_construct_request(self) -> None:
        handler = GeckoVersionProtocolHandler.request(1, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>AVERS\x01</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoVersionProtocolHandler.response(
            (1, 2, 3), (4, 5, 6), parms=PARMS
        )
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>SVERS\x00\x01\x02\x03\x00\x04\x05\x06</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoVersionProtocolHandler()
        self.assertTrue(handler.can_handle(b"AVERS", PARMS))
        self.assertTrue(handler.can_handle(b"SVERS", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_handle_request(self) -> None:
        handler = GeckoVersionProtocolHandler()
        handler.handle(b"AVERS\x01", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)

    def test_recv_handle_response(self) -> None:
        handler = GeckoVersionProtocolHandler()
        handler.handle(b"SVERS\x00\x01\x02\x03\x00\x04\x05\x06", PARMS)
        self.assertTrue(handler.should_remove_handler)
        self.assertEqual(handler.en_build, 1)
        self.assertEqual(handler.en_major, 2)
        self.assertEqual(handler.en_minor, 3)
        self.assertEqual(handler.co_build, 4)
        self.assertEqual(handler.co_major, 5)
        self.assertEqual(handler.co_minor, 6)


if __name__ == "__main__":
    unittest.main()



"""GeckoLib utilities."""

from .async_command import AsyncCmd
from .cui import CUI
from .shell import GeckoShell
from .simulator import GeckoSimulator
from .snapshot import GeckoSnapshot

__all__ = ["CUI", "AsyncCmd", "GeckoShell", "GeckoSimulator", "GeckoSnapshot"]



"""Gecko Pumps."""

from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING

from geckolib.automation.power import GeckoPower
from geckolib.const import GeckoConstants

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade
    from geckolib.driver.accessor import GeckoStructAccessor

_LOGGER = logging.getLogger(__name__)


class GeckoPump(GeckoPower):
    """Pumps are similar to switches, but might have variable speeds too."""

    class PumpType(Enum):
        """Types of pump."""

        NONE = 0
        SINGLE_SPEED = 1
        TWO_SPEED = 2
        VARIABLE_SPEED = 3

    def __init__(self, facade: GeckoAsyncFacade, name: str, key: str) -> None:
        """Initialize the pump class."""
        super().__init__(facade, name, key)
        self.device_class = GeckoConstants.DEVICE_CLASS_PUMP
        self.pump_type = GeckoPump.PumpType.NONE
        self._state_accessor: GeckoStructAccessor | None = None

        if key in facade.spa.struct.connections:
            self.pump_type = GeckoPump.PumpType.SINGLE_SPEED
        if key.upper() in facade.spa.struct.connections:
            self.pump_type = GeckoPump.PumpType.SINGLE_SPEED

        # Look for P1H, P2H etc
        if f"{key}H" in facade.spa.struct.connections:
            self.pump_type = GeckoPump.PumpType.SINGLE_SPEED

        # Look for BLO
        if f"{key}O" in facade.spa.struct.connections:
            self.pump_type = GeckoPump.PumpType.SINGLE_SPEED

        # Look for P1L, P2L etc
        if f"{key}L" in facade.spa.struct.connections:
            if self.pump_type == GeckoPump.PumpType.NONE:
                self.pump_type = GeckoPump.PumpType.SINGLE_SPEED
            else:
                self.pump_type = GeckoPump.PumpType.TWO_SPEED

        udkey = f"Ud{key}"

        if key == "P1" and "Pump1AsVSP" in facade.spa.accessors:
            p1vsp = facade.spa.accessors["Pump1AsVSP"]
            if p1vsp.value is True:
                self.pump_type = GeckoPump.PumpType.VARIABLE_SPEED
                udkey = "UdVSP1"

        if key == "P3" and "Pump3AsVSP" in facade.spa.accessors:
            p3vsp = facade.spa.accessors["Pump3AsVSP"]
            if p3vsp.value is True:
                self.pump_type = GeckoPump.PumpType.VARIABLE_SPEED
                udkey = "UdVSP3"

        if self.pump_type != GeckoPump.PumpType.NONE and udkey in facade.spa.accessors:
            self._state_accessor = facade.spa.accessors[udkey]
            self._state_accessor.watch(self._on_change)
            self.set_availability(is_available=True)

    @property
    def is_on(self) -> bool:
        """Return True if the device is running, False otherwise."""
        if self._state_accessor is None:
            return False
        if (
            self._state_accessor.accessor_type
            == GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE
        ):
            return self._state_accessor.value
        if (
            self._state_accessor.accessor_type
            == GeckoConstants.SPA_PACK_STRUCT_BYTE_TYPE
        ):
            return self._state_accessor.value > 0
        return self._state_accessor.value != "OFF"

    @property
    def is_variable_speed(self) -> bool:
        """Is this a variable speed pump."""
        return self.pump_type == GeckoPump.PumpType.VARIABLE_SPEED

    @property
    def modes(self) -> list[str]:
        """Get the pump modes."""
        if self._state_accessor is None or not self._state_accessor.items:
            return []
        return self._state_accessor.items

    @property
    def mode(self) -> str:
        """Get the pump mode."""
        if self._state_accessor is None:
            return "NA"
        return self._state_accessor.value

    @property
    def percentage(self) -> int:
        """Get the speed percentage."""
        return self._state_accessor.value

    async def async_turn_on(
        self,
        percentage: int | None = None,
        preset_mode: str | None = None,
    ) -> None:
        """Turn on the pump."""
        _LOGGER.info("Turn on pump %s with %s amd %s", self, percentage, preset_mode)
        if self.pump_type == GeckoPump.PumpType.VARIABLE_SPEED:
            if percentage is None:
                percentage = 100
            await self._state_accessor.async_set_value(percentage)
        else:
            if preset_mode is None:
                preset_mode = "HI"
            # Check that it is in the list of modes, otherwise we use "ON"
            if (
                self._state_accessor.items is not None
                and preset_mode not in self._state_accessor.items
            ):
                preset_mode = "ON"
            await self._state_accessor.async_set_value(preset_mode)

    async def async_turn_off(self) -> None:
        """Turn off the pump."""
        if self.pump_type == GeckoPump.PumpType.VARIABLE_SPEED:
            await self._state_accessor.async_set_value(0)
        else:
            await self._state_accessor.async_set_value("OFF")

    async def async_set_mode(self, mode: str) -> None:
        """Set the mode."""
        if self._state_accessor is not None:
            _LOGGER.debug("%s async set mode %s", self.name, mode)
            await self._state_accessor.async_set_value(mode)

    @property
    def state(self) -> str:
        """Get state."""
        if self.pump_type == GeckoPump.PumpType.VARIABLE_SPEED and self.mode == 0:
            return "OFF"
        return f"{self.mode}"

    def __str__(self) -> str:
        """Stringize class."""
        return f"{self.name}: {self.state}"

    @property
    def monitor(self) -> str:
        """Get monitor string."""
        return f"{self.key}: {self.state}"



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoHelloProtocolHandler,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoHelloHandler(unittest.TestCase):
    """Test the GeckoHelloProtocol classes."""

    def setUp(self) -> None:
        """Set up the class."""
        self.handler = GeckoHelloProtocolHandler(b"")

    def test_send_broadcast_construct(self) -> None:
        handler = GeckoHelloProtocolHandler.broadcast()
        self.assertEqual(handler.send_bytes, b"<HELLO>1</HELLO>")

    def test_send_client_construct(self) -> None:
        handler = GeckoHelloProtocolHandler.client(b"CLIENT")
        self.assertEqual(handler.send_bytes, b"<HELLO>CLIENT</HELLO>")

    def test_send_response_construct(self) -> None:
        handler = GeckoHelloProtocolHandler.response(b"SPA", "Name")
        self.assertEqual(handler.send_bytes, b"<HELLO>SPA|Name</HELLO>")

    def test_recv_can_handle(self) -> None:
        self.assertTrue(self.handler.can_handle(b"<HELLO></HELLO>", ()))
        self.assertFalse(self.handler.can_handle(b"<HELLO></HELLO", ()))
        self.assertFalse(self.handler.can_handle(b"<HELLO></HELLO> ", ()))
        self.assertFalse(self.handler.can_handle(b"<GOODBYE>", ()))

    def test_recv_broadcast(self) -> None:
        self.assertFalse(self.handler.handle(b"<HELLO>1</HELLO>", ()))
        self.assertTrue(self.handler.was_broadcast_discovery)
        self.assertIsNone(self.handler._client_identifier)
        self.assertIsNone(self.handler._spa_identifier)
        self.assertIsNone(self.handler._spa_name)

    def test_recv_client_ios(self) -> None:
        self.assertFalse(self.handler.handle(b"<HELLO>IOSCLIENT</HELLO>", ()))
        self.assertFalse(self.handler.was_broadcast_discovery)
        self.assertEqual(self.handler.client_identifier, b"IOSCLIENT")
        self.assertIsNone(self.handler._spa_identifier)
        self.assertIsNone(self.handler._spa_name)

    def test_recv_client_android(self) -> None:
        self.assertFalse(self.handler.handle(b"<HELLO>ANDCLIENT</HELLO>", ()))
        self.assertFalse(self.handler.was_broadcast_discovery)
        self.assertEqual(self.handler.client_identifier, b"ANDCLIENT")
        self.assertIsNone(self.handler._spa_identifier)
        self.assertIsNone(self.handler._spa_name)

    @unittest.expectedFailure
    def test_recv_client_unknown(self) -> None:
        self.handler.handle(b"<HELLO>UNKCLIENT</HELLO>", ())

    def test_recv_response(self) -> None:
        self.assertFalse(self.handler.handle(b"<HELLO>SPA|Name</HELLO>", ()))
        self.assertFalse(self.handler.was_broadcast_discovery)
        self.assertIsNone(self.handler._client_identifier)
        self.assertEqual(self.handler.spa_identifier, b"SPA")
        self.assertEqual(self.handler.spa_name, "Name")

    def test_recv_can_handle_multiple(self) -> None:
        self.test_recv_response()
        self.test_recv_client_android()
        self.test_recv_broadcast()


if __name__ == "__main__":
    unittest.main()



"""Unit tests for the Async locator class."""  # noqa: INP001

from unittest import IsolatedAsyncioTestCase

from context import GeckoAsyncLocator


class TestAsyncLocator(IsolatedAsyncioTestCase):
    """Test the GeckoAsyncLocator class."""

    def setUp(self) -> None:
        self.locator = GeckoAsyncLocator(None, None)

    async def asyncSetUp(self) -> None:
        pass

    async def asyncTearDown(self) -> None:
        pass

    def tearDown(self) -> None:
        del self.locator

    #####################################################

    def test_on_construct(self) -> None:
        self.assertIsNone(self.locator.spas)
        self.assertFalse(self.locator.is_running)
        self.assertEqual(self.locator.age, 0)
        self.assertFalse(self.locator.has_had_enough_time)



"""Energy base class."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from geckolib.automation.base import GeckoAutomationFacadeBase

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade


class GeckoPower(GeckoAutomationFacadeBase):
    """Base class for energy counting objects."""

    def __init__(self, facade: GeckoAsyncFacade, name: str, key: str) -> None:
        """Initialize the energy base class."""
        super().__init__(facade, name, key)
        self.power: float = 0.0

    @property
    @abstractmethod
    def is_on(self) -> bool:
        """Determine if this device is on or not."""

    @property
    def current_power(self) -> float:
        """Get powewr value for this device if it is on."""
        return self.power if self.is_on else 0.0



"""Constants Class."""


class GeckoConstants:
    """
    All the constants in one place.

    GeckoConstants is a literal class so that we can program in a mostly DRY fashion,
    for example a filename or url would be present as would values that could be changed
    and might be difficult to discover inline, but some constants that form part of a
    functions documented behaviour might not be here, e.g. 'rw' as a parameter to open()
    """

    INTOUCH2_PORT = 10022
    # Maximum time to wait for full connection for a responding spa
    CONNECTION_TIMEOUT_IN_SECONDS = 45
    # Async Locator sleep times
    ASYNC_LOCATOR_BROADCAST_SLEEP = 1
    # Simulator min time between responses to broadcast <HELLO>
    SIMULATOR_MIN_TIME_BETWEEN_HELLO_BROADCAST_RESPONSES = 1

    CONNECTION_STEP_PAUSE_IN_SECONDS = None  # Time between connection steps
    MAX_RF_ERRORS_BEFORE_HALT = 50

    BROADCAST_ADDRESS = "<broadcast>"
    MESSAGE_ENCODING = "latin1"
    FORMAT_CLIENT_IDENTIFIER = "IOS{0}"

    # XPaths and attributes for SpaPackStruct
    SPA_PACK_STRUCT_BYTE_TYPE = "Byte"
    SPA_PACK_STRUCT_WORD_TYPE = "Word"
    SPA_PACK_STRUCT_TIME_TYPE = "Time"
    SPA_PACK_STRUCT_BOOL_TYPE = "Bool"
    SPA_PACK_STRUCT_ENUM_TYPE = "Enum"
    # Accessor keys for SpaPackStruct
    KEY_PACK_TYPE = "PackType"
    KEY_PACK_CONFIG_ID = "PackConfID"
    KEY_PACK_CONFIG_REV = "PackConfRev"
    KEY_PACK_CONFIG_REL = "PackConfRel"
    KEY_PACK_CORE_ID = "PackCoreID"
    KEY_PACK_CORE_REV = "PackCoreRev"
    KEY_PACK_CORE_REL = "PackCoreRel"
    KEY_CONFIG_NUMBER = "ConfigNumber"
    KEY_TEMP_UNITS = "TempUnits"
    KEY_RH_WATER_TEMP = "RhWaterTemp"
    KEY_SETPOINT_G = "SetpointG"
    KEY_REAL_SETPOINT_G = "RealSetPointG"
    KEY_DISPLAYED_TEMP_G = "DisplayedTempG"
    KEY_HEATING = "Heating"
    KEY_COOLINGDOWN = "CoolingDown"
    KEY_TEMP_NOT_VALID = "TempNotValid"
    KEY_PUMP_1 = "P1"
    KEY_PUMP_2 = "P2"
    KEY_PUMP_3 = "P3"
    KEY_PUMP_4 = "P4"
    KEY_PUMP_5 = "P5"
    KEY_BLOWER = "BL"
    KEY_WATERFALL = "Waterfall"
    KEY_CIRCULATING_PUMP = "CP"
    KEY_OZONE = "O3"
    KEY_SMW_ACTIVE = "SwmActive"
    KEY_SWM_RISK = "SwmRisk"
    KEY_FILTER_PURGE = "Purge"
    KEY_FILTER_CLEAN = "Clean"
    KEY_ECON_ACTIVE = "EconActive"
    KEY_INGRID_DETECTED = "InGridDetected"
    KEY_MODBUS_HEATPUMP_DETECTED = "ModbusHeatPumpDetected"
    KEY_COOLZONE_DETECTED = "CoolZoneDetected"
    KEY_COOLZONE_MODE = "CoolZoneMode"
    KEY_LOCKMODE = "LockMode"
    KEY_STANDBY = "QuietState"

    KEY_MRSTEAM_USER_MODE = "UserMode"
    KEY_MRSTEAM_USER_AROMA = "UserAroma"
    KEY_MRSTEAM_USER_CHROMA = "UserChroma"
    KEY_MRSTEAM_USER_SETPOINT_G = "UserSetpointG"
    KEY_MRSTEAM_USER_RUNTIME = "UserRuntime"

    KEY_MRSTEAM_HEATER_OUTPUT = "HeaterOutput"
    KEY_MRSTEAM_AROMA_OUTPUT = "AromaOutput"
    KEY_MRSTEAM_CHROMA_OUTPUT = "ChromaOutput"
    KEY_MRSTEAM_REMAINING_RUNTIME = "RemainingRuntime"
    KEY_MRSTEAM_MAX_RUNTIME = "MaxRuntime"

    KEY_BAIN_POWER_STATE = "PowerState"
    KEY_BAIN_GEYSAIR = "UserGeysair"
    KEY_BAIN_HEATER1 = "UserHeater1"
    KEY_BAIN_HEATER2 = "UserHeater2"
    KEY_BAIN_DURATION = "UserBathTime"
    KEY_BAIN_INTENSITY = "UserBlowerIntensity"
    KEY_BAIN_DRYING_CYCLE = "UserDryingCycle"
    KEY_BAIN_DRYING_CYCLE_DELAY = "UserDryingDelay"
    KEY_BAIN_DRYING_CYCLE_HOUR = "UserDryingHour"
    KEY_BAIN_DRYING_CYCLE_MINUTE = "UserDryingMinute"
    KEY_BAIN_CHROMA = "UserChroma"
    KEY_BAIN_HAS_CHROMO = "LL_Chromo"

    KEY_MIN_SETPOINT_G = "MinSetpointG"
    KEY_MAX_SETPOINT_G = "MaxSetpointG"

    KEY_USER_DEMAND_LIGHT = "UdLi"

    EXCEPTION_MESSAGE_NOT_WRITABLE = (
        "Cannot set value for {0}. This status array item doesn't allow writing"
    )

    # Water heater status
    WATER_HEATER_HEATING = "Heating"
    WATER_HEATER_COOLING = "Cooling"
    WATER_HEATER_IDLE = "Idle"

    # Gecko keypad constants
    KEYPAD_PUMP_1 = 1
    KEYPAD_PUMP_2 = 2
    KEYPAD_PUMP_3 = 3
    KEYPAD_PUMP_4 = 4
    KEYPAD_PUMP_5 = 5
    KEYPAD_BLOWER = 6
    KEYPAD_LIGHT = 16
    KEYPAD_LIGHT_120 = 17
    KEYPAD_UP = 21
    KEYPAD_DOWN = 22
    KEYPAD_WATERFALL = 23
    KEYPAD_AUX = 24
    KEYPAD_ECOMODE = 0

    # Pack outputs
    DEVICE_CLASS_PUMP = "PUMP"
    DEVICE_CLASS_BLOWER = "BLOWER"
    DEVICE_CLASS_LIGHT = "LIGHT"
    DEVICE_CLASS_SWITCH = "SWITCH"
    DEVICE_CLASS_OTHER = "OTHER"

    ECON_ACTIVE_DESCRIPTION = "Economy Mode"

    BINARY_SENSORS = [  # noqa: RUF012
        (
            "Circulating Pump",
            KEY_CIRCULATING_PUMP,
            DEVICE_CLASS_PUMP,
        ),
        ("Pump Run", "PumpRun", DEVICE_CLASS_PUMP),
        ("Ozone", KEY_OZONE, DEVICE_CLASS_OTHER),
        ("Smart Winter Mode:Active", KEY_SMW_ACTIVE, DEVICE_CLASS_OTHER),
        ("Filter Status:Clean", KEY_FILTER_CLEAN, DEVICE_CLASS_OTHER),
        ("Filter Status:Purge", KEY_FILTER_PURGE, DEVICE_CLASS_OTHER),
        ("Heating", KEY_HEATING, DEVICE_CLASS_OTHER),
    ]

    SENSORS = [("Smart Winter Mode:Risk", KEY_SWM_RISK)]  # noqa: RUF012

    WATERCARE_MODE = (
        AwayFromHome,
        Standard,
        EnergySaving,
        SuperEnergySaving,
        Weekender,
    ) = range(5)

    WATERCARE_MODE_STRING = [  # noqa: RUF012
        "Away From Home",
        "Standard",
        "Energy Saving",
        "Super Energy Saving",
        "Weekender",
    ]

    PACK_NAME_ADJUSTMENTS = {  # noqa: RUF012
        "MrSt": "MrSteam",
        "MASI": "MAS-IBC-32K",
    }



"""Simulator action class."""

# ruff: noqa: N802

import logging
from typing import Any

_LOGGER = logging.getLogger(__name__)


class GeckoSimulatorAction:
    """Define the action for various simulator events."""

    def set_connections(self, connections: list[str]) -> None:
        """Set the internal connections variable."""
        self._connections = connections

    ############################################################################
    #
    #               The accessors that we want to interact with
    #
    # User demands
    UdP1: Any
    UdVSP1: Any
    UdP2: Any
    UdP3: Any
    UdVSP3: Any
    UdP4: Any
    UdP5: Any
    UdBL: Any
    UdWaterfall: Any
    UdAux: Any
    UdLi: Any
    UdL120: Any
    UdQuietTime: Any
    SetpointG: Any

    # Mr.Steam
    UserMode: Any
    HeaterOutput: Any
    UserAroma: Any
    AromaOutput: Any
    UserChroma: Any
    ChromaOutput: Any

    # Support accessors
    P1: Any
    P2: Any
    P3: Any
    P4: Any
    P5: Any
    BL: Any
    Waterfall: Any
    L120: Any
    CP: Any
    QuietState: Any
    EconActive: Any

    # Water heater
    DisplayedTempG: Any
    RealSetPointG: Any
    MSTR_HEATER: Any
    Heating: Any
    HeaterPump: Any
    TempUnits: Any
    UserRuntime: Any
    RemainingRuntime: Any

    # Utility accessors
    CheckFlo: Any
    UdPumpTime: Any
    UdWaterFallTime: Any
    UdAuxTime: Any
    UdLightTime: Any
    UdL120Time: Any

    Pump1AsVSP: Any
    Pump3AsVSP: Any
    WaterfallAsCP: Any

    _connections: list[str]

    ############################################################################
    #
    #               The keypad buttons we want to handle
    #
    def on_KEYPAD_PUMP_1(self) -> None:
        """Handle keypad pump 1."""
        if self.Pump1AsVSP:
            self.UdVSP1 = self._next_pump_value(self.UdVSP1, vsp=True)
        else:
            self.UdP1 = self._next_pump_value(self.UdP1, "P1L")

    def on_KEYPAD_PUMP_2(self) -> None:
        """Handle keypad pump 2."""
        self.UdP2 = self._next_pump_value(self.UdP2, "P2L")

    def on_KEYPAD_PUMP_3(self) -> None:
        """Handle keypad pump 3."""
        if self.Pump3AsVSP:
            self.UdVSP3 = self._next_pump_value(self.UdVSP3, vsp=True)
        else:
            self.UdP3 = self._next_pump_value(self.UdP3, "P3L")

    def on_KEYPAD_PUMP_4(self) -> None:
        """Handle keypad pump 4."""
        self.UdP4 = self._next_pump_value(self.UdP4, "P4L")

    def on_KEYPAD_PUMP_5(self) -> None:
        """Handle keypad pump 5."""
        self.UdP5 = self._next_pump_value(self.UdP5, "P5L")

    def on_KEYPAD_BLOWER(self) -> None:
        """Handle blower keypad button."""
        self.UdBL = "ON" if self.UdBL == "OFF" else "OFF"

    def on_KEYPAD_WATERFALL(self) -> None:
        """Handle waterfall button."""
        self.UdWaterfall = "ON" if self.UdWaterfall == "OFF" else "OFF"

    def on_KEYPAD_AUX(self) -> None:
        """Handle aux button."""
        self.UdAux = "ON" if self.UdAux == "OFF" else "OFF"

    def _next_pump_value(
        self, ud: str | int, low_conn: str = "NA", *, vsp: bool = False
    ) -> str | int:
        if vsp:
            return (ud + 20) % 120
        two_speed = low_conn in self._connections
        if two_speed:
            return "LO" if ud == "OFF" else "HI" if ud == "LO" else "OFF"
        return "HI" if ud == "OFF" else "OFF"

    def on_KEYPAD_LIGHT(self) -> None:
        """Handle keypad for light."""
        self.UdLi = "HI" if self.UdLi == "OFF" else "OFF"

    def on_KEYPAD_LIGHT_120(self) -> None:
        """Handle keypad for light 2."""
        self.UdL120 = "ON" if self.UdL120 == "OFF" else "OFF"

    def on_KEYPAD_UP(self) -> None:
        """Handle keypad temp up."""
        self.SetpointG += 0.5 if self.TempUnits == "C" else 1.0

    def on_KEYPAD_DOWN(self) -> None:
        """Handle keypad temp down."""
        self.SetpointG -= 0.5 if self.TempUnits == "C" else 1.0

    ############################################################################
    #
    #               The pack properties that we want to watch
    #
    def on_UdP1(self) -> None:
        """Handle UdP1 changes."""
        if self.UdP1 == "HI":
            self.P1 = "HIGH"

        elif self.UdP1 == "LO":
            self.P1 = "LOW"

        else:
            self.P1 = "OFF"

    def on_UdP2(self) -> None:
        """Handle UdP2 changes."""
        if self.UdP2 == "HI":
            self.P2 = "HIGH"

        elif self.UdP2 == "LO":
            self.P2 = "LOW"

        else:
            self.P2 = "OFF"

    def on_UdP3(self) -> None:
        """Handle UdP3 changes."""
        if self.UdP3 == "HI":
            self.P3 = "HIGH"

        elif self.UdP3 == "LO":
            self.P3 = "LOW"

        else:
            self.P3 = "OFF"

    def on_UdP4(self) -> None:
        """Handle UdP4 changes."""
        if self.UdP4 == "HI":
            self.P4 = "HIGH"

        elif self.UdP4 == "LO":
            self.P4 = "LOW"

        else:
            self.P4 = "OFF"

    def on_UdP5(self) -> None:
        """Handle UdP5 changes."""
        if self.UdP5 == "HI":
            self.P5 = "HIGH"

        elif self.UdP5 == "LO":
            self.P5 = "LOW"

        else:
            self.P5 = "OFF"

    def on_UdBL(self) -> None:
        """Handle UdBL changes."""
        self.BL = self.UdBL

    def on_UdWaterfall(self) -> None:
        """Handle UdWaterfall changes."""
        self.Waterfall = self.UdWaterfall

    def on_UdAux(self) -> None:
        """Handle UdAux changes."""
        if self.UdAux == "ON":
            self.UdAuxTime = 20
        else:
            self.UdAuxTime = 0

    def on_UdLi(self) -> None:
        """Handle UdLi changes."""
        self.UdLightTime = 60 if self.UdLi == "HI" else 0

    def on_UdL120(self) -> None:
        """Handle UdL120 changes."""
        self.UdL120Time = 60 if self.UdL120 == "ON" else 0
        self.L120 = "ON" if self.UdL120 == "ON" else "OFF"

    def on_P1(self) -> None:
        """Handle changes to P1."""
        self._pump_helper()

    def on_P2(self) -> None:
        """Handle changes to P2."""
        self._pump_helper()

    def on_P3(self) -> None:
        """Handle changes to P3."""
        self._pump_helper()

    def on_P4(self) -> None:
        """Handle changes to P4."""
        self._pump_helper()

    def on_P5(self) -> None:
        """Handle changes to P5."""
        self._pump_helper()

    def on_BL(self) -> None:
        """Handle changes to BL."""
        self._pump_helper()

    def on_Waterfall(self) -> None:
        """Handle changes to Waterfall."""
        if self.Waterfall == "ON":
            self.UdWaterFallTime = 30
        else:
            self.UdWaterFallTime = 0

    def on_QuietState(self) -> None:
        """Handle quiet state."""
        if self.QuietState == "OFF":
            self.UdQuietTime = 29
            self.UdP1 = "OFF"
            self.UdVSP1 = 0
            self.UdP2 = "OFF"
            self.UdP3 = "OFF"
            self.UdP4 = "OFF"
            self.UdP5 = "OFF"
            self.UdBL = "OFF"
            self.UdWaterfall = "OFF"
            self.UdAux = "OFF"
            self.UdVSP3 = 0
        else:
            self.UdQuietTime = 0

    def on_SetpointG(self) -> None:
        """Handle changes to setpoint."""
        self._temp_helper()

    def on_UserMode(self) -> None:
        """Handle Mr.Steam user mode."""
        self.HeaterOutput = self.UserMode == "ON"
        if self.UserRuntime == 0:
            self.UserRuntime = 2
        if self.HeaterOutput:
            self.RemainingRuntime = self.UserRuntime

    def on_UserAroma(self) -> None:
        """Handle Mr.Steam aroma mode."""
        self.AromaOutput = self.UserAroma == "ON"

    def on_UserChroma(self) -> None:
        """Handle Mr.Steam chroma mode."""
        self.ChromaOutput = self.UserChroma == "ON"

    def on_UserRuntime(self) -> None:
        """Handle Mr.Steam user runtime change."""
        self.RemainingRuntime = self.UserRuntime

    ############################################################################
    #
    #               Timer routines
    #

    def every_second(self, _total_seconds: int) -> None:
        """Perform action every second."""
        self._pump_helper()
        self._temp_helper()

    def every_minute(self, _total_minutes: int) -> None:
        """Perform action every minute."""

    def every_hour(self, _total_hours: int) -> None:
        """Perform action every minute."""

    ############################################################################
    #
    #               Support and help routines
    #

    def _pump_helper(self) -> None:
        if all(
            [
                self.P1 == "OFF",
                self.P2 == "OFF",
                self.P3 == "OFF",
                self.P4 == "OFF",
                self.P5 == "OFF",
                self.BL == "OFF",
            ]
        ):
            self.CheckFlo = False
            self.UdPumpTime = 0
        else:
            self.CheckFlo = True
            self.UdPumpTime = 15

    def _temp_helper(self) -> None:
        if any(
            [
                self.RealSetPointG is None,
                self.SetpointG is None,
                self.DisplayedTempG is None,
            ]
        ):
            return

        # Set the real setpoint temperature
        self.RealSetPointG = self.SetpointG

        # Handle heating and cooling
        if self.DisplayedTempG >= self.RealSetPointG:
            # We're at temperature, so we turn off heaters etc
            if self.MSTR_HEATER is not None:
                self.MSTR_HEATER = "OFF"
            if self.Heating is not None:
                if isinstance(self.Heating, bool):
                    self.Heating = False
                else:
                    self.Heating = ""

            self.CP = "OFF"

        elif self.DisplayedTempG < self.RealSetPointG - 0.5:
            # Below 0.5C of the setpoint, turn stuff on
            if self.MSTR_HEATER is not None:
                self.MSTR_HEATER = "ON"
            if self.Heating is not None:
                if isinstance(self.Heating, bool):
                    self.Heating = True
                else:
                    self.Heating = "Heating"

            self.CP = "ON"

    @property
    def is_heating(self) -> bool:
        """Determine if heating is being done."""
        if self.MSTR_HEATER is not None:
            return self.MSTR_HEATER == "ON"
        if self.Heating is not None:
            if isinstance(self.Heating, bool):
                return self.Heating
            return self.Heating != ""
        return False

    def check_thermodynamics(self, _elapsed_time: int) -> None:
        """Check thermodynamics for the spa."""
        # If there is no heating, then the water temperature drops
        delta: float = 0.2 if self.TempUnits == "F" else 0.1
        if self.DisplayedTempG is not None:
            if self.is_heating:
                self.DisplayedTempG += delta
            else:
                self.DisplayedTempG -= delta



"""Async Facade to hide implementation details."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any

from geckolib.automation.bainultra import BainUltra
from geckolib.automation.base import GeckoAutomationBase
from geckolib.automation.bubblegen import GeckoBubbleGenerator
from geckolib.automation.heatpump import GeckoHeatPump
from geckolib.automation.ingrid import GeckoInGrid
from geckolib.automation.inmix import GeckoInMix
from geckolib.automation.lockmode import GeckoLockMode
from geckolib.automation.mrsteam import MrSteam
from geckolib.automation.waterfall import GeckoWaterfall
from geckolib.config import GeckoConfig, config_sleep, set_config_mode
from geckolib.const import GeckoConstants
from geckolib.driver import Observable

from .blower import GeckoBlower
from .heater import GeckoWaterHeater
from .keypad import GeckoKeypad
from .light import GeckoLight, GeckoLightL120, GeckoLightLi
from .pump import GeckoPump
from .reminders import GeckoReminders
from .sensors import GeckoBinarySensor, GeckoErrorSensor, GeckoSensor, GeckoSensorBase
from .switch import GeckoStandby, GeckoSwitch
from .watercare import GeckoWaterCare

if TYPE_CHECKING:
    from geckolib.async_spa import GeckoAsyncSpa
    from geckolib.async_taskman import GeckoAsyncTaskMan
    from geckolib.automation.power import GeckoPower

_LOGGER = logging.getLogger(__name__)


class GeckoAsyncFacade(Observable):
    """Async facade."""

    class SpaInUseSensor(GeckoAutomationBase):
        """Sensor with idle/active state."""

        def __init__(self, facade: GeckoAsyncFacade) -> None:
            """Initializd the in-use sensor."""
            super().__init__(facade.unique_id, "Spa In Use", facade.name, "INUSE")
            self._facade: GeckoAsyncFacade = facade
            self._in_use: bool = False

        @property
        def is_on(self) -> bool:
            """Determine if the sensor is on or not."""
            state = self.state
            if isinstance(state, bool):
                return state
            if state == "":
                return False
            return state != "OFF"

        @property
        def state(self) -> bool:
            """The state of the sensor."""
            return self._in_use

        @property
        def device_class(self) -> str | None:
            """The device class."""
            return None

        @property
        def unit_of_measurement(self) -> str | None:
            """The unit of measurement for the sensor, or None."""
            return None

        def set_in_use(self, *, in_use: bool) -> None:
            """Set the internal flag."""
            self._in_use = in_use
            self._on_change()

        def __repr__(self) -> str:
            """Get string representation of the class."""
            return f"{self.name}: {self.state}"

    def __init__(
        self, spa: GeckoAsyncSpa, taskman: GeckoAsyncTaskMan, **_kwargs: str
    ) -> None:
        """
        Initialize the facade.

        This is a thin automation-friendly client on the spa implementation.
        """
        Observable.__init__(self)

        self._spa: GeckoAsyncSpa = spa
        self._taskman: GeckoAsyncTaskMan = taskman

        spa.struct.build_connections()

        # Declare all the items that a spa might have. If they are
        # available for this configuration, they will be marked as such.
        self.pump_1: GeckoPump = GeckoPump(self, "Pump 1", "P1")
        self.pump_2: GeckoPump = GeckoPump(self, "Pump 2", "P2")
        self.pump_3: GeckoPump = GeckoPump(self, "Pump 3", "P3")
        self.pump_4: GeckoPump = GeckoPump(self, "Pump 4", "P4")
        self.pump_5: GeckoPump = GeckoPump(self, "Pump 5", "P5")

        self.blower: GeckoBlower = GeckoBlower(self)
        self.waterfall: GeckoWaterfall = GeckoWaterfall(self)
        self.bubblegenerator: GeckoBubbleGenerator = GeckoBubbleGenerator(self)

        self.light: GeckoLightLi = GeckoLightLi(self)
        self.light2: GeckoLightL120 = GeckoLightL120(self)

        self._heatpump: GeckoHeatPump = GeckoHeatPump(self)
        self._ingrid: GeckoInGrid = GeckoInGrid(self)
        self._lockmode: GeckoLockMode = GeckoLockMode(self)

        self._inmix: GeckoInMix = GeckoInMix(self)

        self._mrsteam: MrSteam = MrSteam(self)
        self._bainultra: BainUltra = BainUltra(self)

        #################

        # Declare all the class members
        self._sensors: list[GeckoSensorBase] = []
        self._binary_sensors: list[GeckoBinarySensor] = []
        self._pumps: list[GeckoPump] = []
        self._blowers: list[GeckoBlower] = []
        self._lights: list[GeckoLight] = []
        self._ecomode: GeckoSwitch | None = None
        self._standby: GeckoStandby | None = None

        # Build the automation items
        self._reminders_manager = GeckoReminders(self)
        self._water_heater = GeckoWaterHeater(self)
        self._water_care = GeckoWaterCare(self)
        self._keypad = GeckoKeypad(self)
        self._scan_outputs()
        # Install change notifications
        for device in self.all_automation_devices:
            device.watch(self._on_change)
        # And notifications for active/idle items
        for device in self.all_config_change_devices:
            device.watch(self._on_config_device_change)

        self._taskman.add_task(self._facade_update(), "Facade update", "FACADE")
        self._ready = asyncio.Event()

        self._spa_in_use_sensor = GeckoAsyncFacade.SpaInUseSensor(self)

    async def disconnect(self) -> None:
        """Disconnect the facade."""
        try:
            _LOGGER.debug("Disconnect facade")
            self._taskman.cancel_key_tasks("FACADE")
            for device in self.all_automation_devices:
                device.unwatch_all()
        except Exception:
            _LOGGER.exception("During facade disconnect")

    def _on_config_device_change(self, *_args: Any) -> None:
        in_use = False
        for device in self.all_config_change_devices:
            if device.is_on:
                in_use = True
        set_config_mode(active=in_use)
        self._spa_in_use_sensor.set_in_use(in_use=in_use)

    async def _facade_update(self) -> None:
        _LOGGER.debug("Facade update task started")
        try:
            do_check = True
            while True:
                # Only want to do the protocol work if timeout, otherwise we run
                # the risk of swamping the spa when we're making operational changes
                wait_time = GeckoConfig.FACADE_UPDATE_FREQUENCY_IN_SECONDS
                if not self._spa.is_responding_to_pings:
                    wait_time = GeckoConfig.PING_FREQUENCY_IN_SECONDS
                elif do_check:
                    if self._water_care.is_available:
                        _LOGGER.debug("Facade update: Get watercare mode")
                        self._water_care.change_watercare_mode(
                            await self._spa.async_get_watercare_mode()
                        )
                    if self._reminders_manager.is_available:
                        _LOGGER.debug("Facade update: Get reminders")
                        self._reminders_manager.change_reminders(
                            await self._spa.async_get_reminders()
                        )
                    self._on_config_device_change()

                    # After we've been round here at least once, we're ready
                    self._ready.set()

                do_check = await config_sleep(wait_time, "Async facade update loop")
                _LOGGER.debug("Facade update, wait over. do_check=%s", f"{do_check}")

        except asyncio.CancelledError:
            _LOGGER.debug("Facade update loop cancelled")
            raise
        except Exception:
            _LOGGER.exception("Facade update loop caught execption")
            raise

    async def wait_for_one_update(self) -> None:
        """Wait until the first update has occurred."""
        await self._ready.wait()

    def _scan_outputs(self) -> None:
        """Scan the spa outputs to decide what user options are available."""
        if self._spa is None:
            return

        self._pumps: list[GeckoPump] = [
            pump
            for pump in [
                self.pump_1,
                self.pump_2,
                self.pump_3,
                self.pump_4,
                self.pump_5,
                self.waterfall,
                self.bubblegenerator,
            ]
            if pump.is_available
        ]

        self._blowers = [blower for blower in [self.blower] if blower.is_available]

        self._lights = [
            light for light in [self.light, self.light2] if light.is_available
        ]

        #######

        self._sensors = [
            GeckoSensor(self, sensor[0], self._spa.accessors[sensor[1]])
            for sensor in GeckoConstants.SENSORS
            if sensor[1] in self._spa.accessors
        ]

        if self.pump_1.is_available:
            self._sensors.append(
                GeckoSensor(self, "Pump 1 State", self.spa.accessors["P1"])
            )
        if self.pump_2.is_available:
            self._sensors.append(
                GeckoSensor(self, "Pump 2 State", self.spa.accessors["P2"])
            )
        if self.pump_3.is_available:
            self._sensors.append(
                GeckoSensor(self, "Pump 3 State", self.spa.accessors["P3"])
            )
        if self.pump_4.is_available:
            self._sensors.append(
                GeckoSensor(self, "Pump 4 State", self.spa.accessors["P4"])
            )
        if self.pump_5.is_available:
            self._sensors.append(
                GeckoSensor(self, "Pump 5 State", self.spa.accessors["P5"])
            )
        if self.blower.is_available:
            self._sensors.append(
                GeckoSensor(self, "Blower State", self.spa.accessors["BL"])
            )
        if self.waterfall.is_available:
            self._sensors.append(
                GeckoSensor(self, "Waterfall State", self.spa.accessors["Waterfall"])
            )

        self._binary_sensors = [
            GeckoBinarySensor(
                self, binary_sensor[0], self._spa.accessors[binary_sensor[1]]
            )
            for binary_sensor in GeckoConstants.BINARY_SENSORS
            if binary_sensor[1] in self._spa.accessors
        ]

        self._error_sensor = GeckoErrorSensor(self)

        if GeckoConstants.KEY_ECON_ACTIVE in self._spa.accessors:
            self._ecomode = GeckoSwitch(
                self,
                GeckoConstants.KEY_ECON_ACTIVE,
                (
                    GeckoConstants.ECON_ACTIVE_DESCRIPTION,
                    GeckoConstants.KEY_ECON_ACTIVE,
                    GeckoConstants.DEVICE_CLASS_SWITCH,
                ),
            )

        if GeckoConstants.KEY_STANDBY in self._spa.accessors:
            self._standby = GeckoStandby(self)

    @property
    def unique_id(self) -> str:
        """A unique id for the facade."""
        return self._taskman.unique_id

    @property
    def name(self) -> str:
        """Get the spa name."""
        return self._taskman.spa_name

    @property
    def spa(self) -> GeckoAsyncSpa:
        """Get the spa implementation instance."""
        return self._spa

    @property
    def mrsteam(self) -> MrSteam:
        """Get the Mr.Steam object."""
        return self._mrsteam

    @property
    def bainultra(self) -> BainUltra:
        """Get the Bain Ultra object."""
        return self._bainultra

    @property
    def reminders_manager(self) -> GeckoReminders:
        """Get the reminders handler."""
        return self._reminders_manager

    @property
    def water_heater(self) -> GeckoWaterHeater:
        """Get the water heater handler."""
        return self._water_heater

    @property
    def water_care(self) -> GeckoWaterCare:
        """Get the water care handler."""
        return self._water_care

    @property
    def keypad(self) -> GeckoKeypad:
        """Get the keypad handler."""
        return self._keypad

    @property
    def pumps(self) -> list[GeckoPump]:
        """Get the pumps list."""
        return self._pumps

    @property
    def blowers(self) -> list[GeckoBlower]:
        """Get the blowers list."""
        return self._blowers

    @property
    def lights(self) -> list[GeckoLight]:
        """Get the lights list."""
        return self._lights

    @property
    def switches(self) -> list[GeckoSwitch]:
        """Get the switches."""
        if self.mrsteam.is_available:
            return self.mrsteam.switches
        if self.bainultra.is_available:
            return self.bainultra.switches
        switches = []
        if self.eco_mode is not None:
            switches.append(self.eco_mode)
        if self.standby is not None:
            switches.append(self.standby)
        return switches

    @property
    def sensors(self) -> list[GeckoSensorBase]:
        """Get the sensor list."""
        return self._sensors

    @property
    def binary_sensors(self) -> list[GeckoBinarySensor]:
        """Get the binary sensor list."""
        return self._binary_sensors

    @property
    def error_sensor(self) -> GeckoErrorSensor:
        """Get the error sensor."""
        return self._error_sensor

    @property
    def eco_mode(self) -> GeckoSwitch | None:
        """Get the Eco Mode switch if available."""
        return self._ecomode

    @property
    def heatpump(self) -> GeckoHeatPump:
        """Get the heat pump if available."""
        return self._heatpump

    @property
    def ingrid(self) -> GeckoInGrid:
        """Get the inGrid handler if available."""
        return self._ingrid

    @property
    def lockmode(self) -> GeckoLockMode:
        """Get the lockmode handler."""
        return self._lockmode

    @property
    def standby(self) -> GeckoStandby | None:
        """Get the standby switch if available."""
        return self._standby

    @property
    def inmix(self) -> GeckoInMix:
        """Get the inMix handler."""
        return self._inmix

    @property
    def spa_in_use_sensor(self) -> GeckoAsyncFacade.SpaInUseSensor:
        """Get the spa in use sensor."""
        return self._spa_in_use_sensor

    @property
    def all_user_devices(self) -> list[GeckoAutomationBase]:
        """Get all the user controllable devices as a list."""
        return list(self._pumps + self._blowers + self._lights)

    @property
    def all_config_change_devices(self) -> list[GeckoPower]:
        """Get all devices that can cause config change."""
        return list(self._pumps + self._blowers + self._lights)

    @property
    def all_automation_devices(self) -> list[GeckoAutomationBase]:
        """Get all the automation devices as a list."""
        extras = []
        extras.extend(self.all_user_devices)
        extras.extend(self.sensors)
        extras.extend(self.binary_sensors)
        extras.append(self.water_heater)
        extras.append(self.water_care)
        extras.append(self.reminders_manager)
        extras.append(self.keypad)
        if self.eco_mode is not None:
            extras.append(self.eco_mode)
        if self.heatpump is not None:
            extras.append(self.heatpump)
        if self.ingrid is not None:
            extras.append(self.ingrid)
        if self.mrsteam.is_available:
            extras.append(self.mrsteam)
        if self.bainultra.is_available:
            extras.append(self.bainultra)
        return extras

    def get_device(self, key: str) -> GeckoAutomationBase | None:
        """Get an automation device from the key, or None if not found."""
        for device in self.all_automation_devices:
            if device.key == key:
                return device
        return None

    @property
    def devices(self) -> list[str]:
        """
        Get a list of automation device keys.

        Keys can be passed to get_device to find the specific device.
        """
        return [device.key for device in self.all_automation_devices]

    @property
    def taskmanager(self) -> GeckoAsyncTaskMan:
        """The facade's task manager."""
        return self._taskman



"""Unit tests for the actual snapshots."""  # noqa: INP001

import pathlib
from unittest import IsolatedAsyncioTestCase, main

from context import (
    GeckoAsyncFacade,
    GeckoAsyncStructure,
    GeckoAsyncTaskMan,
    GeckoSnapshot,
)

from geckolib.driver.accessor import GeckoStructAccessor


class GeckoAsyncSpa:
    """Mock spa for testing."""

    def __init__(self, snapshotfile: str) -> None:
        """Initialize the class."""
        self.struct = GeckoAsyncStructure(None)

        cwd = pathlib.Path(__file__).parent.resolve()
        snapshots = GeckoSnapshot.parse_log_file(f"{cwd}/{snapshotfile}")
        assert len(snapshots) == 1

        snapshot = snapshots[0]
        self.struct.replace_status_block_segment(0, snapshot.bytes)

        # Attempt to get config and log classes
        self.plateform_key = snapshot.packtype.lower()
        self.config_version = snapshot.config_version
        self.log_version = snapshot.log_version

    @property
    def accessors(self) -> list[GeckoStructAccessor]:
        """Get the accessors."""
        return self.struct.accessors

    @property
    def is_responding_to_pings(self) -> bool:
        """Is the remote end responding to pings."""
        return True

    async def async_init(self) -> None:
        """Init async."""
        await self.struct.load_pack_class(self.plateform_key)
        await self.struct.load_config_module(self.config_version)
        await self.struct.load_log_module(self.log_version)
        await self.struct.check_for_accessories()
        self.struct.build_accessors()


class TestSnapshots(IsolatedAsyncioTestCase):
    """Test all the snapshots."""

    def setUp(self) -> None:
        self.taskman = GeckoAsyncTaskMan()

    async def asyncSetUp(self) -> None:
        await self.taskman.__aenter__()

    async def asyncTearDown(self) -> None:
        await self.taskman.__aexit__(None)

    def tearDown(self) -> None:
        del self.taskman

    async def build_facade(self, snapshotfile: str) -> GeckoAsyncFacade:
        spa = GeckoAsyncSpa(snapshotfile)
        await spa.async_init()
        return GeckoAsyncFacade(spa, self.taskman)

    async def test_default(self) -> None:
        facade = await self.build_facade("snapshots/default.snapshot")
        self.assertTrue(facade.pump_1.is_available)
        self.assertTrue(facade.pump_2.is_available)
        self.assertFalse(facade.pump_3.is_available)
        self.assertTrue(facade.blower.is_available)
        self.assertTrue(facade.light.is_available)
        self.assertFalse(facade.light2.is_available)
        self.assertFalse(facade.pumps[0].is_on)
        self.assertFalse(facade.pumps[1].is_on)
        self.assertFalse(facade.blowers[0].is_on)
        self.assertFalse(facade.lights[0].is_on)
        self.assertEqual(39.0, facade.water_heater.current_temperature)
        self.assertEqual("°C", facade.water_heater.temperature_unit)

    async def test_inYT_Pump1Low(self) -> None:  # noqa: N802
        facade = await self.build_facade(
            "snapshots/inYT-Pump1Lo-2020-12-13 11_19_35.snapshot"
        )
        self.assertTrue(facade.pump_1.is_available)
        self.assertTrue(facade.pump_2.is_available)
        self.assertTrue(facade.light.is_available)
        self.assertEqual("LO", facade.pumps[0].mode)
        self.assertEqual("OFF", facade.pumps[1].mode)
        self.assertTrue(facade.pumps[0].is_on)
        self.assertFalse(facade.pumps[1].is_on)
        self.assertTrue(facade.lights[0].is_on)
        self.assertEqual(70.0, facade.water_heater.current_temperature)
        self.assertEqual("°F", facade.water_heater.temperature_unit)

    async def test_inYT_Pump1High(self) -> None:  # noqa: N802
        facade = await self.build_facade(
            "snapshots/inYT-Pump1Hi-2020-12-13 11_19_35.snapshot"
        )
        self.assertTrue(facade.pump_1.is_available)
        self.assertTrue(facade.pump_2.is_available)
        self.assertTrue(facade.light.is_available)
        self.assertEqual("HI", facade.pumps[0].mode)
        self.assertEqual("OFF", facade.pumps[1].mode)
        self.assertTrue(facade.pumps[0].is_on)
        self.assertFalse(facade.pumps[1].is_on)
        self.assertTrue(facade.lights[0].is_on)
        self.assertEqual(70.0, facade.water_heater.current_temperature)
        self.assertEqual("°F", facade.water_heater.temperature_unit)

    async def test_inYT_WaterFall_On(self) -> None:  # noqa: N802
        facade = await self.build_facade(
            "snapshots/inYT-waterfall on-2020-10-23 18_01_30.snapshot"
        )
        self.assertTrue(facade.pump_1.is_available)
        self.assertTrue(facade.pump_2.is_available)
        self.assertTrue(facade.waterfall.is_available)
        self.assertTrue(facade.light.is_available)
        self.assertEqual("OFF", facade.pumps[0].mode)
        self.assertEqual("OFF", facade.pumps[1].mode)
        self.assertEqual("ON", facade.pumps[2].mode)
        self.assertFalse(facade.pumps[0].is_on)
        self.assertFalse(facade.pumps[1].is_on)
        self.assertFalse(facade.lights[0].is_on)
        self.assertEqual(37.0, facade.water_heater.current_temperature)
        self.assertEqual("°C", facade.water_heater.temperature_unit)

    async def test_inYJ_All_Off(self) -> None:  # noqa: N802
        facade = await self.build_facade(
            "snapshots/inYJ-All off-2020-12-18 11_24_09.snapshot"
        )
        self.assertTrue(facade.pump_1.is_available)
        self.assertTrue(facade.light.is_available)
        self.assertEqual("OFF", facade.pumps[0].mode)
        self.assertFalse(facade.pumps[0].is_on)
        self.assertFalse(facade.lights[0].is_on)
        self.assertEqual(37.0, facade.water_heater.current_temperature)
        self.assertEqual("°C", facade.water_heater.temperature_unit)


if __name__ == "__main__":
    main()



"""Automation interface support classes."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.driver import Observable

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoAutomationBase(Observable):
    """Base of all the automation helper classes."""

    def __init__(self, unique_id: str, name: str, parent_name: str, key: str) -> None:
        """Initialize the base class."""
        super().__init__()
        self._unique_id: str = unique_id
        self._name: str = name
        self._parent_name: str = parent_name
        self._key: str = key
        self._is_available: bool = False
        self.mapping = {}
        self.reverse = {}

    def set_mapping(self, mapping: dict) -> None:
        """Set the mapping and the reverse mapping helpers."""
        self.mapping = mapping
        self.reverse = {v: k for k, v in self.mapping.items()}

    @property
    def name(self) -> str:
        """All automation items have a name."""
        return self._name

    @property
    def parent_name(self) -> str:
        """All automation items have a parent that has a name."""
        return self._parent_name

    @property
    def key(self) -> str:
        """Key into the spa pack."""
        return self._key

    @property
    def unique_id(self) -> str:
        """A unique id for the property."""
        return f"{self._unique_id}-{self._key}"

    @property
    def parent_unique_id(self) -> str:
        """The parent unique ID."""
        return f"{self._unique_id}"

    @property
    def monitor(self) -> str:
        """An abbreviated string for the monitor process in the shell."""
        return f"{self}"

    @property
    def is_available(self) -> bool:
        """Get the availability of this item."""
        return self._is_available

    def set_availability(self, *, is_available: bool) -> None:
        """Set the availability of this item."""
        self._is_available = is_available

    def __repr__(self) -> str:
        """Get string representation."""
        return (
            f"{super().__repr__()}(name={self.name},"
            f" parent={self.parent_name} key={self.key})"
        )


class GeckoAutomationFacadeBase(GeckoAutomationBase):
    """Base of all the automation helper classes from the Facade."""

    def __init__(self, facade: GeckoAsyncFacade, name: str, key: str) -> None:
        """Initialize the facade base class."""
        super().__init__(facade.unique_id, name, facade.name, key)
        self._facade = facade
        self._spa = facade.spa

    @property
    def facade(self) -> GeckoAsyncFacade:
        """Return the facade that is associated with this automation object."""
        return self._facade



"""Unit tests for the AsyncPeekableQueue ."""  # noqa: INP001

# ruff: noqa: E501

from geckolib.driver.async_peekablequeue import AsyncPeekableQueue


class TestAsyncPeekableQueue:
    """Tests for async peekable queue."""

    def test_empty_on_construction(self) -> None:
        q = AsyncPeekableQueue()
        assert not q.is_data_available
        assert not q
        assert len(q) == 0

    def test_with_state(self) -> None:
        q = AsyncPeekableQueue()
        q.push(1)
        q.push(2)
        assert q.is_data_available
        assert q
        assert len(q) == 2
        one = q.pop()
        assert q.is_data_available
        assert one == 1
        assert q
        assert len(q) == 1
        two = q.pop()
        assert two == 2
        assert not q.is_data_available
        assert not q
        assert len(q) == 0

    def test_command_priority(self) -> None:
        q = AsyncPeekableQueue()
        q.push(1)
        q.push(2)
        q.push(3)
        q.push_command(4)
        assert len(q) == 4
        assert q.peek() == 4
        four = q.pop()
        assert four == 4
        assert q.is_data_available
        assert q.peek() == 1



"""Automation keypad backlight class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .select import GeckoSelect

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoKeypadBacklight(GeckoSelect):
    """The keypad backlight value."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the backlight class."""
        super().__init__(facade, "Keypad Backlight", "KeypadBacklightColor")
        # Set of mappings of constants to UI options
        self.set_mapping(
            {
                "OFF": "Off",
                "RED": "Red",
                "GREEN": "Green",
                "YELLOW": "Yellow",
                "BLUE": "Blue",
                "MAGENTA": "Magenta",
                "CYAN": "Cyan",
                "WHITE": "White",
            }
        )

    @property
    def state(self) -> str:
        """Get the current state via the mapping."""
        return self.mapping[self._accessor.value]

    async def async_set_state(self, new_state: str) -> None:
        """Set the state of the select entity."""
        if new_state in self.reverse:
            new_state = self.reverse[new_state]
        await self._accessor.async_set_value(new_state)

    @property
    def states(self) -> list[str]:
        """Get the possible states."""
        return list(self.mapping.values())



"""Gecko STATU/STATV/STATQ/STATP handlers."""

from __future__ import annotations

import logging
import struct
from typing import TYPE_CHECKING, Any

from geckolib.config import GeckoConfig

from .packet import GeckoPacketProtocolHandler

if TYPE_CHECKING:
    from geckolib.driver.async_udp_protocol import GeckoAsyncUdpProtocol

STATU_VERB = b"STATU"
STATV_VERB = b"STATV"
STATQ_VERB = b"STATQ"
STATP_VERB = b"STATP"

REQUEST_FORMAT = ">BHH"
RESPONSE_FORMAT = ">BBB"

_LOGGER = logging.getLogger(__name__)


class GeckoStatusBlockProtocolHandler(GeckoPacketProtocolHandler):
    """Handle full status block."""

    @staticmethod
    def request(
        seq: int, start: int, length: int, **kwargs: Any
    ) -> GeckoStatusBlockProtocolHandler:
        """Generate a range request."""
        return GeckoStatusBlockProtocolHandler(
            start=start,
            content=b"".join(
                [STATU_VERB, struct.pack(REQUEST_FORMAT, seq, start, length)]
            ),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS * 3,
            on_retry_failed=GeckoPacketProtocolHandler.default_retry_failed_handler,
            **kwargs,
        )

    @staticmethod
    def full_request(seq: int, **kwargs: Any) -> GeckoStatusBlockProtocolHandler:
        """Generate a full request."""
        return GeckoStatusBlockProtocolHandler.request(seq, 0, 1024, **kwargs)

    @staticmethod
    def response(
        index: int, next_index: int, block: bytes, **kwargs: Any
    ) -> GeckoStatusBlockProtocolHandler:
        """Generate a response."""
        return GeckoStatusBlockProtocolHandler(
            start=0,
            content=b"".join(
                [
                    STATV_VERB,
                    struct.pack(RESPONSE_FORMAT, index, next_index, len(block)),
                    block,
                ]
            ),
            **kwargs,
        )

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the class."""
        super().__init__(**kwargs)
        self.start = kwargs.get("start")
        self.sequence = self.length = self.next = self.data = None

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle this."""
        return received_bytes.startswith((STATU_VERB, STATV_VERB))

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the verb."""
        remainder = received_bytes[5:]
        if received_bytes.startswith(STATU_VERB):
            self.sequence, self.start, self.length = struct.unpack(
                REQUEST_FORMAT, remainder
            )
            return  # Stay in the handler list

        # Otherwise must be STATV
        self.sequence, self.next, self.length = struct.unpack(
            RESPONSE_FORMAT, remainder[0:3]
        )
        self.data = remainder[3 : self.length + 3]
        _LOGGER.debug(
            "Status block segment # %d (then #%d) length %d, %r",
            self.sequence,
            self.next,
            self.length,
            self.data,
        )

    def __repr__(self) -> str:
        """Get a string representation."""
        return (
            f"{super().__repr__()}(seq={self.sequence},start={self.start},"
            f"length={self.length},next={self.next},data={self.data})"
        )


class GeckoAsyncPartialStatusBlockProtocolHandler(GeckoPacketProtocolHandler):
    """Async partial status block handler."""

    def __init__(self, protocol: GeckoAsyncUdpProtocol, **kwargs: Any) -> None:
        """Initialize the class."""
        super().__init__(**kwargs)
        self._protocol: GeckoAsyncUdpProtocol = protocol
        self.changes = []

    @staticmethod
    def report_changes(
        protocol: GeckoAsyncUdpProtocol, changes: list[tuple], **kwargs: Any
    ) -> GeckoAsyncPartialStatusBlockProtocolHandler:
        """Report changes as a list of change tuples, (pos, data)."""
        change_bin = [(struct.pack(">H", change[0]), change[1]) for change in changes]
        change_list = [item for change in change_bin for item in change]
        pack_data = b"".join(
            [STATP_VERB, struct.pack(">B", len(changes)), *change_list]
        )
        if len(pack_data) % 2 != 0:
            msg = "Pack data must be even length, check change lengths"
            raise RuntimeError(msg)
        return GeckoAsyncPartialStatusBlockProtocolHandler(
            protocol,
            content=pack_data,
            **kwargs,
        )

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Decide what we can handle."""
        return received_bytes.startswith((STATQ_VERB, STATP_VERB))

    def handle(self, received_bytes: bytes, sender: tuple) -> None:
        """Handle the block."""

    async def async_handle(self, received_bytes: bytes, sender: tuple) -> None:
        """Handle the block."""
        remainder = received_bytes[5:]
        if received_bytes.startswith(STATQ_VERB):
            (self.sequence,) = struct.unpack(">B", remainder)
            return  # Stay in the handler list

        # Otherwise must be STATP
        self._protocol.queue_send(
            GeckoPacketProtocolHandler(
                content=b"".join(
                    [
                        STATQ_VERB,
                        struct.pack(
                            ">B",
                            self._protocol.get_and_increment_sequence_counter(
                                command=True
                            ),
                        ),
                    ]
                ),
                parms=sender,
            ),
        )

        change_count = struct.unpack(">B", remainder[0:1])[0]
        self.changes = []
        for i in range(change_count):
            pos = struct.unpack(">H", remainder[1 + (i * 4) : 3 + (i * 4)])[0]
            self.changes.append((pos, remainder[3 + (i * 4) : 5 + (i * 4)]))



"""Bain Ultra class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.automation.heater import GeckoWaterHeaterAbstract
from geckolib.automation.number import GeckoNumber
from geckolib.automation.select import GeckoSelect
from geckolib.automation.switch import GeckoSwitch
from geckolib.const import GeckoConstants

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class BainUltra(GeckoWaterHeaterAbstract):
    """Bain Ultra support class."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the Bain Ultra class."""
        super().__init__(facade, "Bath", "BAIN")
        if not facade.spa.struct.is_bain_ultra:
            return

        self.set_availability(is_available=True)

        self.onoff = GeckoSwitch(
            facade,
            GeckoConstants.KEY_BAIN_POWER_STATE,
            (
                "On/Off",
                GeckoConstants.KEY_BAIN_POWER_STATE,
                GeckoConstants.DEVICE_CLASS_SWITCH,
            ),
        )
        self.onoff.watch(self._on_change)
        # The spa pack didn't set Read/Write on the power property, but the BU client
        # wrote to it anyway, so we will too...
        self.facade.spa.accessors[GeckoConstants.KEY_BAIN_POWER_STATE].set_read_write(
            "ALL"
        )

        self.geysair = GeckoSwitch(
            facade,
            GeckoConstants.KEY_BAIN_GEYSAIR,
            (
                "Geysair",
                GeckoConstants.KEY_BAIN_GEYSAIR,
                GeckoConstants.DEVICE_CLASS_BLOWER,
            ),
        )
        self.geysair.watch(self._on_change)
        # The spa pack didn't set Read/Write on the geysair property, but the BU client
        # wrote to it anyway, so we will too...
        self.facade.spa.accessors[GeckoConstants.KEY_BAIN_GEYSAIR].set_read_write("ALL")

        self.backrest = GeckoSelect(
            facade, "Heated Backrest", GeckoConstants.KEY_BAIN_HEATER1
        )
        self.backrest.watch(self._on_change)
        self.backrest2 = GeckoSelect(
            facade, "2nd Heated Backrest", GeckoConstants.KEY_BAIN_HEATER2
        )
        self.backrest2.watch(self._on_change)

        self.bath_runtime = GeckoNumber(
            facade, "Duration", GeckoConstants.KEY_BAIN_DURATION, "min"
        )
        self.bath_runtime.native_max_value = 30.0
        self.bath_runtime.watch(self._on_change)
        self.bath_intensity = GeckoNumber(
            facade, "Intensity", GeckoConstants.KEY_BAIN_INTENSITY, "%"
        )
        self.bath_intensity.watch(self._on_change)

        self.drying_cycle = GeckoSelect(
            facade, "Drying Cycle", GeckoConstants.KEY_BAIN_DRYING_CYCLE
        )
        self.drying_cycle.watch(self._on_change)
        self.drying_cycle_delay = GeckoNumber(
            facade,
            "Drying Cycle Delay",
            GeckoConstants.KEY_BAIN_DRYING_CYCLE_DELAY,
            "min",
        )
        self.drying_cycle_delay.native_step = 5.0
        self.drying_cycle_delay.watch(self._on_change)

        self.drying_cycle_hour = GeckoNumber(
            facade, "Drying Cycle Hour", GeckoConstants.KEY_BAIN_DRYING_CYCLE_HOUR, "hr"
        )
        self.drying_cycle_hour.native_max_value = 23.0
        self.drying_cycle_hour.watch(self._on_change)

        self.drying_cycle_minute = GeckoNumber(
            facade,
            "Drying Cycle Minute",
            GeckoConstants.KEY_BAIN_DRYING_CYCLE_MINUTE,
            "min",
        )
        self.drying_cycle_minute.native_max_value = 59.0
        self.drying_cycle_minute.native_step = 5.0
        self.drying_cycle_minute.watch(self._on_change)

        self.chroma = GeckoSelect(facade, "Chroma", GeckoConstants.KEY_BAIN_CHROMA)
        has_chromo = self.facade.spa.accessors[GeckoConstants.KEY_BAIN_HAS_CHROMO]
        self.chroma.set_availability(is_available=has_chromo.value == "DELUXE")

    @property
    def switches(self) -> list[GeckoSwitch]:
        """Get a list of all the switches."""
        if self.is_available:
            return [self.onoff, self.geysair]
        return []

    @property
    def selects(self) -> list[GeckoSelect]:
        """Get a list of all the selects."""
        if self.is_available:
            return [self.backrest, self.backrest2]
        return []

    @property
    def current_operation(self) -> str:
        """Get the current operation."""
        if not self.onoff.is_on:
            return "Turned off"
        return f"Running, {self.bath_runtime} remaining"
